from unittest import TestSuite

from . import test_splitter
from . import test_waveguide
from . import test_helpers
from . import test_resonator
from . import test_chip


def test_suite():
    suite = TestSuite()
    suite.addTest(test_splitter.test_suite())
    suite.addTest(test_waveguide.test_suite())
    suite.addTest(test_helpers.test_suite())
    suite.addTest(test_resonator.test_suite())
    suite.addTest(test_chip.test_suite())

    return suite



from enum import Enum


class _Configuration(object):
    DefaultDatatype = Enum('DefaultDatatype', 'zero aslayer dose_factor')

    _datatype = DefaultDatatype.aslayer
    _dose_factor = 1.0
    _point_limit = 4000
    _point_limit_line = 4000

    MACHINE_DEFAULTS = {
        'unlimited': {
            'datatype_policy': DefaultDatatype.zero,
            'dose_factor': 1.,
            'point_limit': None,
            'point_limit_line': None
        },
        'JEOL': {
            'datatype_policy': DefaultDatatype.aslayer,
            'dose_factor': 1.,
            'point_limit': 4000,
            'point_limit_line': 4000
        },
        'E-Line': {
            'datatype_policy': DefaultDatatype.dose_factor,
            'dose_factor': 1.,
            'point_limit': 4000,
            'point_limit_line': 4000
        },
        'conservative': {
            'datatype_policy': DefaultDatatype.zero,
            'dose_factor': 1.,
            'point_limit': 200,
            'point_limit_line': 200
        }
    }

    @property
    def datatype_policy(self):
        return self._datatype

    @datatype_policy.setter
    def datatype_policy(self, policy):
        assert policy in self.DefaultDatatype
        self._datatype = policy

    @property
    def dose_factor(self):
        return self._dose_factor

    @dose_factor.setter
    def dose_factor(self, dose_factor):
        self._dose_factor = float(dose_factor)

    @property
    def point_limit(self):
        return self._point_limit

    @point_limit.setter
    def point_limit(self, max_points):
        assert type(max_points) == int and max_points > 5
        self._point_limit = max_points

    @property
    def point_limit_line(self):
        return self._point_limit

    @point_limit_line.setter
    def point_limit_line(self, max_points_line):
        assert type(max_points_line) == int and max_points_line > 5
        self._point_limit_line = max_points_line

    def set_target_profile(self, profile_name):
        assert profile_name in self.MACHINE_DEFAULTS, 'Profile name must be one of %s' % self.MACHINE_DEFAULTS.keys()

        for key, value in self.MACHINE_DEFAULTS[profile_name].items():
            self.__setattr__(key, value)


import matplotlib.pyplot as plt
import matplotlib.patches
import matplotlib.text

import numpy as np

def save_as_image(obj, file_path, resolution=1., antialiased=True, include_text=False,
                  ylim=(None, None), xlim=(None, None)):
    """
    Save any gdsCAD object as an image.

    You can either use a rasterized file format such as png but also formats such as SVG or PDF.

    :param file_path:
    :param obj: Any gdsCAD object
    :param path: Path of the image file.
    :param resolution: Rasterization resolution in GDSII units.
    :param antialiased: Whether to use a anti-aliasing or not.
    :param include_text: Include text in the output.
    :param ylim: Tuple of (min_x, max_x) to export.
    :param xlim: Tuple of (min_y, max_y) to export.
    """

    # For vector graphics, map 1um to {resolution} mm instead of inch.
    is_vector = file_path.split('.')[-1] in ('svg', 'svgz', 'eps', 'ps', 'emf', 'pdf')
    scale = 5/127. if is_vector else 1.

    dpi = 1./resolution

    fig = plt.figure(frameon=False)
    fig.patch.set_visible(False)

    ax = fig.add_subplot(1, 1, 1)
    ax.set_aspect('equal')
    #ax.margins(0.1)
    ax.axis('off')

    artists=obj.artist()
    for a in artists:
        if not include_text and isinstance(a, matplotlib.text.Text):
            continue

        if hasattr(a, 'set_antialiased'):
            a.set_antialiased(antialiased)

        a.set_transform(a.get_transform() + ax.transData)
        if isinstance(a, matplotlib.patches.Patch):
            ax.add_patch(a)
        elif isinstance(a, matplotlib.lines.Line2D):
            ax.add_line(a)
        else:
            ax.add_artist(a)

    plt.subplots_adjust(left=0., right=1., top=1., bottom=0.)

    # Autoscale, then change the axis limits and read back what is actually displayed
    ax.autoscale(True, tight=True)
    ax.set_xlim(*xlim)
    ax.set_ylim(*ylim)
    actual_ylim, actual_xlim = ax.get_ylim(), ax.get_xlim()


    fig.set_size_inches(np.asarray((actual_xlim[1]-actual_xlim[0], actual_ylim[1]-actual_ylim[0]))*scale)
    fig.set_dpi(dpi)
    fig.savefig(file_path, transparent=True, dpi=dpi)
    plt.close(fig)


def add_jeol_helpers(cell):
    def extract_special(cell):
        return {
            'cells': [extract_special(ref.ref_cell) for ref in cell.references],
            'elements': [e for e in cell.elements if (hasattr(e, 'is_special') and e.is_special)]
        }
    from pprint import pprint
    pprint(extract_special(cell))

def _example():
    import gdsCAD.core
    from math import pi
    from gdshelpers.geometry import convert_to_gdscad
    from gdshelpers.parts.waveguide import Waveguide
    from gdshelpers.parts.coupler import GratingCoupler


    left_coupler = GratingCoupler.make_traditional_coupler_from_database([0, 0], 1, 'sn330', 1550)
    wg = Waveguide.make_at_port(left_coupler.port)
    wg.add_straight_segment(length=10)
    wg.add_bend(-pi/2, radius=50)
    wg.add_straight_segment(length=150)
    wg.add_bend(-pi/2, radius=50)
    wg.add_straight_segment(length=10)
    right_coupler = GratingCoupler.make_traditional_coupler_from_database_at_port(wg.current_port, 'sn330', 1550)

    cell = gdsCAD.core.Cell('SIMPLE_DEVICE')
    cell.add(convert_to_gdscad([left_coupler, wg, right_coupler], layer=1))
    cell.add(convert_to_gdscad([left_coupler.get_description_text(side='right'),
                                right_coupler.get_description_text(side='left')], layer=2))

    save_as_image(cell, '/tmp/test.png', resolution=1)
    save_as_image(cell, '/tmp/test.pdf')
    save_as_image(cell, '/tmp/test.svg')

if __name__ == '__main__':
    _example()


import unittest
from gdshelpers.geometry.chip import Cell
from gdshelpers.parts.waveguide import Waveguide
from shapely.geometry import box
from shapely.geometry.point import Point
from shapely.geometry.linestring import LineString
import numpy as np


class DeviceTestCase(unittest.TestCase):
    def test_cell_bounds(self):
        cell = Cell('test_cell')

        wg1 = Waveguide([0, 0], 0, 1)
        wg1.add_straight_segment(10)
        cell.add_to_layer(1, wg1)

        wg2 = Waveguide([40, 0], 0, 1)
        wg2.add_straight_segment(10)
        cell.add_to_layer(2, wg2)

        cell.add_region_layer(100, [1])
        cell.add_region_layer(101, [1, 2])

        self.assertTrue(100 in cell.layer_dict)
        self.assertEqual(cell.bounds, (0.0, -0.5, 50.0, 0.5))

        subcell = Cell('subcell')
        subcell.add_to_layer(3, box(0, 0, 60, 70))

        self.assertEqual(subcell.bounds, (0.0, 0, 60.0, 70))

        # add subcell
        cell.add_cell(subcell, origin=(0, 0))
        self.assertEqual(cell.bounds, (0.0, -0.5, 60.0, 70.))

        subcell2 = Cell('subcell2')
        subcell2.add_to_layer(3, box(0, 0, 60, 70))

        # add subcell at different origin
        cell.add_cell(subcell2, origin=(20, 0))
        self.assertEqual(cell.bounds, (0.0, -0.5, 80.0, 70.))

        # add object to subcell after adding the subcell to cell
        subcell2.add_to_layer(4, box(60, 0, 70, 10))
        self.assertEqual(subcell2.bounds, (0.0, 0, 70.0, 70))
        self.assertEqual(cell.bounds, (0.0, -0.5, 90.0, 70.))

        # check total region layer
        cell.add_region_layer(102)
        self.assertTrue(102 in cell.layer_dict)
        self.assertEqual(cell.get_bounds(layers=[102]), cell.bounds)

        # cell.save('test_cell_bounds')

    def test_empty_cell(self):
        # An empty cell should have 'None' as bounding box
        cell = Cell('test_cell')
        # self.assertEqual(cell.bounds, None)
        cell.add_to_layer(4, box(60, 0, 70, 10))
        self.assertEqual(cell.bounds, (60, 0, 70, 10))

        subcell = Cell('subcell')
        cell.add_cell(subcell, origin=(0, 0))
        # bounding box of cell should still be the the original bbox
        self.assertEqual(cell.bounds, (60, 0, 70, 10))

    def test_point(self):
        # An Point should give a valid bounding box
        cell = Cell('test_point')
        cell.add_to_layer(4, Point(10, 20))
        self.assertEqual(cell.bounds, (10, 20, 10, 20))

    def test_collection(self):
        # An empty collection should lead to 'None' as bounding box
        cell = Cell('test_collection')
        cell.add_to_layer(4, LineString())
        self.assertEqual(cell.bounds, None)

    def test_bounds_rotation(self):
        import numpy.testing as np_testing

        cell1 = Cell('test_cell')
        cell1.add_to_layer(1, box(10, 10, 30, 20))
        np_testing.assert_almost_equal(cell1.bounds, (10, 10, 30, 20))

        cell2 = Cell('root_cell')
        cell2.add_cell(cell1, angle=0.5*np.pi)
        np_testing.assert_almost_equal(cell2.bounds, (-20, 10, -10, 30))

        cell3 = Cell('root_cell')
        cell3.add_cell(cell1, angle=1.5*np.pi)
        np_testing.assert_almost_equal(cell3.bounds, (10, -30, 20, -10))

        cell4 = Cell('root_cell')
        cell4.add_cell(cell1, angle=1*np.pi)
        np_testing.assert_almost_equal(cell4.bounds, (-30, -20, -10, -10))

        # For manually testing the bounds rotation, uncomment the following code and look at the resulting GDS
        """
        cell5 = Cell('root_cell')
        cell5.add_cell(cell1, angle=0.5*np.pi)
        cell5.add_region_layer()
        cell5.save('test_bounds_rotation')
        """

    def test_region_layers(self):
        cell = Cell("test_device")

        wg1 = Waveguide([0, 0], 0, 1)
        wg1.add_straight_segment(10)
        wg1_shapely = wg1.get_shapely_object()
        cell.add_to_layer(1, wg1)

        wg2 = Waveguide([40, 0], 0, 1)
        wg2.add_straight_segment(10)
        wg2_shapely = wg2.get_shapely_object()
        cell.add_to_layer(2, wg2)

        cell.add_region_layer(100)
        cell.add_region_layer(101, [1])
        cell.add_region_layer(102, [1, 2])

        self.assertTrue(100 in cell.layer_dict)
        self.assertTrue(cell.layer_dict[100][0].contains(wg1_shapely))
        self.assertTrue(cell.layer_dict[100][0].contains(wg2_shapely))

        self.assertTrue(101 in cell.layer_dict)
        self.assertTrue(cell.layer_dict[101][0].contains(wg1_shapely))
        self.assertFalse(cell.layer_dict[101][0].contains(wg2_shapely))

        self.assertTrue(102 in cell.layer_dict)
        self.assertTrue(cell.layer_dict[102][0].contains(wg1_shapely))
        self.assertTrue(cell.layer_dict[102][0].contains(wg2_shapely))

    def test_frame(self):
        # Add frame 
        cell = Cell('test_frame')
        cell.add_to_layer(4, box(0, 0, 10, 10))
        cell.add_frame(padding=10, line_width=1., frame_layer=5)
        self.assertEqual(cell.bounds, (-11, -11, 21, 21))
        self.assertEqual(cell.get_bounds(layers=[4]), (0, 0, 10, 10))
        self.assertEqual(cell.get_bounds(layers=[5]), (-11, -11, 21, 21))

        cell.add_frame(padding=10, line_width=1., frame_layer=99, bbox=(0, 0, 2, 3))
        self.assertEqual(cell.get_bounds(layers=[99]), (-11, -11, 13, 14))


def test_suite():
    return unittest.TestLoader().loadTestsFromTestCase(DeviceTestCase)



"""
This module includes a collection of coupler parameters. It is not meant to be used directly but
rather via the :func:`GratingCoupler.make_traditional_coupler_from_database` functions.

Feel free to report your coupler findings for inclusion in this database.
"""

import numpy as np

REFERENCES = {
    ##########
    # The definitions following below were extracted from the lib/coupler
    # directory. I don't know how trustworthy they are. If you have corrections
    # or recommendations just let me know.

    'aln660': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength + 575.9965)/2772.187,
            'grating_ff': 0.72,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_aln660.il',
            'comment': 'Wavelength corrected Dec-20-2010 by Wolfram Pernice, chip R10-6'
        }
    },

    'aln780': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-363.214)/826.55,
            'grating_ff': 0.8,
            'n_gratings': 30,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_aln780.il',
            'comment': 'Wavelength corrected 03-03-2011, WP from chip R32'
        }
    },

    'aln780_ox': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-450.02925)/662.87,
            'grating_ff': 0.76,
            'n_gratings': 30,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_aln780.il',
            'comment': 'Wavelength corrected 03-09-2011, WP from chip R30 + oxide'
        }
    },

    'aln1230': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-280.060566)/1223.47208,
            'grating_ff': 0.8,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_aln1230.il',
            'comment': ''
        }
    },

    'aln1550': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength - 112.258)/1220.68,
            #NOTE: Before changed by Matthias Stegmaier, this was (wavelength - 796.566)/690.4757
            #      As corrected Mar-03-2011 by Wolfram Pernice, chip R29.
            'grating_ff': 0.85,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_aln1550.il',
            'comment': 'Wavelength corrected 2012-08-06 by Matthias Stegmaier, MS003'
        }
    },

    # It is not clear what thick means in this context but this definition also
    # includes the taper length.
    'aln1550_thick': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength - 1045.0715)/454.29771,
            'grating_ff': 0.65,
            'n_gratings': 20,
            'taper_length': 30,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_aln1550.il',
            'comment': 'Wavelength corrected Sep-12-2011 by Wolfram Pernice, chip CP4'
        }
    },

    # NOTE: coupler_bto* was not included, as I don't know what to make of them
    # NOTE: coupler_diamond* was not included, as I don't know what to make of them

    'gan780': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-438.75)/750,
            'grating_ff': 0.72,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_gan780.il',
            'comment': 'Wavelength measured results: 11-01-2010'
        }
    },

    'gan1550': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-1197.31738)/343.49,
            'grating_ff': 0.8,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_gan780.il',
            'comment': 'Wavelength based on Oct2010 GNC1 device results measured 11-01-2010'
        }
    },

    # NOTE: coupler_gan_shg not included, as I don't know what to make of them

    'hs110': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-336.4172)/1710.5,
            'grating_ff': 0.525,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_gan780.il',
            'comment': ''
        }
    },

    'hs110_ox': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-317.1553)/1760.4,
            'grating_ff': 0.5,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_gan780.il',
            'comment': ''
        }
    },

    # NOTE: coupler_opt not included, as I don't know what to make of them
    # NOTE: coupler_shg not included, as I don't know what to make of them

    'si110': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-728.6)/917.0,
            'grating_ff': 0.8,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_si110.il',
            'comment': 'Wavelength newly corrected formula base on Jun2008 device results'
        }
    },

    'si110_ox': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-898.22)/784.3,
            'grating_ff': 0.8,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_si110.il',
            'comment': 'Wavelength newly corrected formula base on Jun2010 device results (C3) with 1200nm PECVD oxide'
        }
    },

    'si110_nb': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-942.8254)/663.2421,
            'grating_ff': 0.8,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_si110.il',
            'comment': 'Wavelength newly corrected formula base on Jun2010 device results (C3) with 1200nm PECVD oxide'
        }
    },

    'si110_ap': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-728.6)/917.0,
            'grating_ff': 0.8,
            'n_gratings': 20,
            'n_ap_gratings': 10,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_si110.il',
            'comment': 'Wavelength newly corrected formula base on Jun2008 device results'
        }
    },

    'si110_nb_ap': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-942.8254)/663.2421,
            'grating_ff': 0.8,
            'n_gratings': 20,
            'n_ap_gratings': 10,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_si110.il',
            'comment': 'Wavelength newly corrected formula 06-24-2011 from device J16'
        }
    },

    'si220': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength*0.715-389.74)/1000.0,
            'grating_ff': 0.8,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_si220.il',
            'comment': 'Wavelength corrected 05/04/2010'
        }
    },

    'sin500': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-198.4740)/1353.5,
            'grating_ff': 0.4,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_sin500.il',
            'comment': ''
        }
    },

    'sn200_va': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-201.34)/955.1,
            'grating_ff': 0.8,
            'n_gratings': 30,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_sn200_va.il',
            'comment': 'Wavelength formula base on Comsol simulation'
        }
    },

    'sn200_vw': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-235.6)/879.1,
            'grating_ff': 0.65,
            'n_gratings': 30,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_sn200_vw.il',
            'comment': 'Wavelength formula base on Comsol simulation'
        }
    },

    'sn330': {
        'parameters': {
            'full_opening_angle': np.deg2rad(40.0),
            'grating_period': lambda wavelength: (wavelength-620.012)/816.57,
            'grating_ff': 0.85,
            'n_gratings': 20,
        },
        'meta': {
            'origin': 'lib/coupler/coupler_sn330',
            'comment': 'Wavelength changed 06/03/2010 by WP'
        }
    },

}


from __future__ import print_function, division

import math
import numpy as np
import shapely.geometry
import shapely.affinity

from gdshelpers.parts import Port
from gdshelpers.geometry import geometric_union
from gdshelpers.parts.waveguide import Waveguide


class CNT(object):
    """
    Creates a CNT source (Electrodes + tapered waveguide) at waveguide port.

    This class implements the electrodes and tapered waveguide for an integrated CNT source.
    It provides the electrodes ports and the connecting waveguide port.
    The Electrode is made up of three part: the round head (defined by the radius) followed by a
    straight fine line of length el_l_fine, followed by a taper to final_width with length el_l_taper and a box with length
    el_l_straight.

    :param origin: Origin of the resonator, which is the start of the input waveguide.
    :param angle: Angle of the input waveguide.
    :param width: Width of the angle waveguide.
    :param gap: Gap between electrode tip and waveguide.
    :param l_taper: length of the waveguide taper used before and after the cnt. i.a. 2*l_taper will be added to the waveguide
    :param w_taper: width of waveguide at the cnts position
    :param el_l_straight: length of electrode part with largest thickness
    :param el_l_taper: length over which 2*el_radius is tapered to el_final_width
    :param el_final_width: largest width of the electrode
    :param el_l_fine: length of small strip to the tip of the electrode
    :param el_radius: radius of electrodes tip
    :param n_points: number of points used to make electrode tip polygon
    """

    def __init__(self, origin, angle, width, gap=0.15, l_taper=10, w_taper=0.9, el_l_straight=8, el_l_taper=2.5,
                 el_final_width=2, el_l_fine=1.4,
                 el_radius=0.1, n_points=128):
        assert gap >= 0, 'Gap must not be negative'
        assert el_radius > 0, 'The electrode radius must not be negative'

        self._origin_port = Port(origin, angle, width)
        cnt_port = self._origin_port.copy().longitudinal_offset(l_taper)
        cnt_port.width = w_taper
        self._cnt_port = cnt_port
        self._out_port = self._origin_port.copy().longitudinal_offset(2 * l_taper)

        self.gap = gap
        self.points = n_points
        self.el_l_taper = el_l_taper
        self.radius = el_radius
        self.el_l_straight = el_l_straight
        self.final_width = el_final_width
        self.el_l_fine = el_l_fine
        self.l_taper = l_taper

        self.el_shift_x = 0
        self.el_shift_y = self._cnt_port.width / 2 + self.gap + self.radius
        self.angle = self._origin_port.angle + math.pi / 2
        self.sample_distance = 0.015

        self.electrodes = None
        self.waveguide = []

        self._make_waveguide()
        self._make_electrodes()

    @classmethod
    def make_at_port(cls, port, **kwargs):
        default_port_param = dict(port.get_parameters())
        default_port_param.update(kwargs)
        del default_port_param['origin']
        del default_port_param['angle']
        del default_port_param['width']

        return cls(port.origin, port.angle, port.width, **default_port_param)

    @property
    def left_electrode_port(self):
        offset = self.el_shift_y + self.el_l_fine + self.el_l_taper + self.el_l_straight
        port = self._cnt_port.parallel_offset(offset)
        port.width = self.final_width
        port.angle = self.angle
        return port

    @property
    def right_electrode_port(self):
        offset = -(self.el_shift_y + self.el_l_fine + self.el_l_taper + self.el_l_straight)
        port = self._cnt_port.parallel_offset(offset)
        port.width = self.final_width
        port.angle = self.angle + math.pi
        return port

    @property
    def out_port(self):
        return self._out_port.copy()

    @property
    def in_port(self):
        return self._origin_port.copy().inverted_direction

    def _make_waveguide(self):
        wg = Waveguide.make_at_port(self._origin_port)
        wg.add_bezier_to_port(self._cnt_port.inverted_direction, 3, sample_distance=self.sample_distance)
        wg.add_bezier_to_port(self._out_port.inverted_direction, 3, sample_distance=self.sample_distance)
        self.waveguide.append(wg)

    def _make_electrodes(self):
        phi = np.linspace(math.pi, 2 * math.pi, self.points)

        circle_points = np.array([self.radius * np.cos(phi), self.radius * np.sin(phi)]).T

        first_part_points = [circle_points[0], (-self.radius, self.el_l_fine), (self.radius, self.el_l_fine),
                             circle_points[-1]]

        taper_points = [first_part_points[1], (-self.final_width / 2, self.el_l_fine + self.el_l_taper),
                        (self.final_width / 2, self.el_l_fine + self.el_l_taper), first_part_points[-2]]

        last_part_points = [taper_points[1],
                            (-self.final_width / 2, self.el_l_fine + self.el_l_taper + self.el_l_straight),
                            (self.final_width / 2, self.el_l_fine + self.el_l_taper + self.el_l_straight),
                            taper_points[-2]]

        cirlce_polygon = shapely.geometry.Polygon(circle_points)
        first_part_polygon = shapely.geometry.Polygon(first_part_points)
        taper_polygon = shapely.geometry.Polygon(taper_points)
        last_part_polygon = shapely.geometry.Polygon(last_part_points)

        polygon = geometric_union([cirlce_polygon, taper_polygon, first_part_polygon, last_part_polygon])

        upper_electrode = shapely.affinity.translate(polygon, self.el_shift_x, self.el_shift_y)
        lower_electrode = shapely.affinity.scale(upper_electrode, xfact=-1, yfact=-1, origin=(0, 0))

        electrode = geometric_union([lower_electrode, upper_electrode])
        electrode = shapely.affinity.rotate(electrode, self._cnt_port.angle, use_radians=True)
        electrode = shapely.affinity.translate(electrode, self._cnt_port.origin[0], self._cnt_port.origin[1])
        self.electrodes = electrode

    def get_shapely_object(self):
        return geometric_union(self.waveguide)


def _cnt_example():
    import gdsCAD.core
    from gdshelpers.geometry import convert_to_gdscad
    from gdshelpers.parts.waveguide import Waveguide

    # photonics
    start_port = Port((0, 0), 0, 1.1)
    wg = Waveguide.make_at_port(start_port)
    wg.add_straight_segment(20)
    cnt_port = wg.current_port
    cnt = CNT.make_at_port(cnt_port, gap=0.4, l_taper=100, w_taper=0.1)
    wg2 = Waveguide.make_at_port(cnt.out_port)
    wg2.add_bend(np.pi / 4, 100)

    cnt2 = CNT.make_at_port(wg2.current_port, gap=0.15)

    union = geometric_union([wg, cnt, wg2, cnt2])

    # electrodes
    el1_l = Waveguide.make_at_port(cnt.left_electrode_port)
    el1_l.add_straight_segment(100, 100)
    el1_r = Waveguide.make_at_port(cnt.right_electrode_port)
    # el1_r.add_straight_segment(100)

    port = cnt2.left_electrode_port
    port.width = 20
    el2_l = Waveguide.make_at_port(port)
    el2_l.add_straight_segment(30)

    el = geometric_union([cnt.electrodes, cnt2.electrodes, el1_l, el1_r, el2_l])

    cell = gdsCAD.core.Cell('test')
    cell.add(convert_to_gdscad(union))
    cell.add(convert_to_gdscad(el, layer=2))
    layout = gdsCAD.core.Layout()
    layout.add(cell)
    layout.show()
    layout.save('CNT_Device_Test.gds')


if __name__ == '__main__':
    _cnt_example()



import imageio
import shapely.geometry
import shapely.ops
import numpy as np


class GdsImage(object):
    """
    An image represented as GDS parts.

    :param origin: Lower left corner of the image.
    :param filename: Filename of the image.
    :param pixel_size: Size of one pixel.
    """

    def __init__(self, origin, filename, pixel_size):
        self.imgdata = imageio.imread(filename, pilmode="1")
        ysize = np.size(self.imgdata, 0)
        self.pixel_size = pixel_size
        self.origin = origin

        self.pixels = list()
        for (y, x), c in np.ndenumerate(self.imgdata):
            if not c:
                box = shapely.geometry.box(x * pixel_size + origin[0],
                                           (ysize - y) * pixel_size + origin[1],
                                           (x + 1) * pixel_size + origin[0],
                                           (ysize - y - 1) * pixel_size + origin[1])
                self.pixels.append(box)

    def get_shapely_object(self):
        return shapely.ops.cascaded_union(self.pixels)


def _example():
    import gdsCAD.core
    from gdshelpers.geometry import convert_to_gdscad

    img = GdsImage([0, 0], "wolfram_monochrome_100.png", 10)

    cell = gdsCAD.core.Cell('TEST_IMAGE')
    cell.add(convert_to_gdscad(img))
    cell.show()
    return cell


if __name__ == '__main__':
    _example()



import unittest
from gdshelpers.parts.resonator import RingResonator


class RingResonatorTestCase(unittest.TestCase):
    def test_asymmetric_gap(self):
        radius = 50.
        wg_width = 1.
        gaps = (0.1, 0.5)

        splitter = RingResonator([0, 0], 0, wg_width, gaps, radius)
        self.assertAlmostEqual(splitter.port.origin[1], 0)
        self.assertAlmostEqual(splitter.opposite_side_port_out.origin[1], 2. * wg_width + gaps[0] + gaps[1] + 2.*radius)
        
    def test_symmetric_gap(self):
        radius = 50.
        wg_width = 1.
        gap = 0.5

        splitter = RingResonator([0, 0], 0, wg_width, gap, radius)
        self.assertAlmostEqual(splitter.port.origin[1], 0)
        self.assertAlmostEqual(splitter.opposite_side_port_out.origin[1], 2. * wg_width + 2. * gap + 2.*radius)
        
    def test_asymmetric_gap_inconsistent_values(self):
        self.assertRaises(ValueError, RingResonator, [0, 0], 0, 1., (0.5, -0.5), 50)
        self.assertRaises(ValueError, RingResonator, [0, 0], 0, 1., (-0.5, 0.5), 50)
        self.assertRaises(ValueError, RingResonator, [0, 0], 0, 1., (1,2,3), 50)


def test_suite():
    return unittest.TestLoader().loadTestsFromTestCase(RingResonatorTestCase)


import numpy as np


class CubicBezierCurve(object):
    def __init__(self, p0, p1, p2, p3):
        self._p0, self._p1, self._p2, self._p3 = [np.asarray(p) for p in (p0, p1, p2, p3)]

    def evaluate(self, t):
        return (1 - t) ** 3 * self._p0[..., None] + 3 * (1 - t) ** 2 * t * self._p1[..., None] \
               + 3 * (1 - t) * t ** 2 * self._p2[..., None] + t ** 3 * self._p3[..., None]

    def evaluate_d1(self, t):
        return 3 * (1 - t) ** 2 * (self._p1[..., None] - self._p0[..., None]) \
               + 6 * (1 - t) * t * (self._p2[..., None] - self._p1[..., None]) \
               + 3 * t ** 2 * (self._p3[..., None] - self._p2[..., None])

    def split(self, t):
        """
        Split the cubic Bezier curve into two new cubic Bezier, both describing one part of the curve.

        :param t:
        :return:
        """
        p1, p2, p3, p4 = self._p0, self._p1, self._p2, self._p3

        p12 = (p2 - p1) * t + p1
        p23 = (p3 - p2) * t + p2
        p34 = (p4 - p3) * t + p3

        p123 = (p23 - p12) * t + p12
        p234 = (p34 - p23) * t + p23

        p1234 = (p234 - p123) * t + p123

        return CubicBezierCurve(p1, p12, p123, p1234), \
               CubicBezierCurve(p1234, p234, p34, p4)


if __name__ == '__main__':
    import matplotlib.pyplot as plt

    cp = [(0, 0), (10, 0), (20, 20), (30, 20)]
    curve = CubicBezierCurve(*cp)
    curve2, curve3 = curve.split(0.1)

    t = np.linspace(0, 1, num=100)

    curve_eval = curve.evaluate(t)
    curve2_eval = curve2.evaluate(t)
    curve3_eval = curve3.evaluate(t)

    plt.figure()
    plt.plot(curve_eval[0, :], curve_eval[1, :])
    plt.plot(curve2_eval[0, :], curve2_eval[1, :])
    plt.plot(curve3_eval[0, :], curve3_eval[1, :])
    plt.show()


import numpy as np
from shapely.affinity import rotate, translate
from shapely.geometry import Polygon

from gdshelpers.geometry import convert_to_gdscad
from gdshelpers.helpers.bezier import CubicBezierCurve
from gdshelpers.geometry import geometric_union
from gdshelpers.parts.port import Port
from gdshelpers.parts.waveguide import Waveguide


class Ntron(object):
    def __init__(self, origin, angle, gate_width_1, gate_width_2, choke_width_1, choke_width_2, choke_length_2,
                 choke_length_3, choke_point_2=(1.0 / 2, 0), choke_point_3=(1, 1.0 / 2), points_per_curve=1000,
                 gate_start=0, gate_point_2=(1.0 / 3, 0), gate_point_3=(2.0 / 3, 1), channel_point_2=(1.0 / 3, 0),
                 channel_point_3=(2.0 / 3, 0), channel_length=2.5, outer_channel_width=0.4, inner_channel_width=0.2,
                 channel_position=0.5, gate_length=0.5, choke_start=0.5, gate_choke_length=0.2):
        """
        Superconducting nanoscale Transistor, for instance to amplify SNSPD signals or creating logical circuits.

        Link: https://pubs.acs.org/doi/10.1021/nl502629x
        All standard parameters are taken from paper above.
        Link Bezier explanation: https://javascript.info/bezier-curve

        :param origin: device origin coordinate (x,y)
        :param angle: device angle (gate to x axis)
        :param gate_width_1: outer width of the gate port
        :param gate_width_2: connection gate-choke, should be same as choke width_1
        :param choke_width_1: outer width of the choke
        :param choke_width_2: Inner choke width, which is the bottleneck and the most important parameter
        :param choke_length_2: length of the gate sided choke bezier shape
        :param choke_length_3: length of the channel sided choke bezier shape
        :param choke_point_2: 2nd bezier point in a rectangle (Look Bezier explanation above)
        :param choke_point_3: 3rd bezier point in a rectangle (Look Bezier explanation above)
        :param points_per_curve: number of of points per Bezier curve used to create a shapely geometry
        :param gate_start: starting x pos of the gate section, should be zero
        :param gate_point_2: 2nd bezier point in a rectangle (Look Bezier explanation above)
        :param gate_point_3: 3rd bezier point in a rectangle (Look Bezier explanation above)
        :param channel_point_2: 2nd bezier point in a rectangle (Look Bezier explanation above)
        :param channel_point_3: 3rd bezier point in a rectangle (Look Bezier explanation above)
        :param channel_length: length of the channel
        :param outer_channel_width: starting width of the channel at the gates
        :param inner_channel_width: end width in the middle of the channel
        :param channel_position: channel position corresponding to the _gate, for 0 choke is on the bottom, for 1 on the top
        :param gate_length: the length of the bezier calculated shape of the gate
        :param choke_start: starting x pos of the choke section, should be equal to gate_length
        :param gate_choke_length:
        """

        self._origin = origin
        self._angle = angle
        self._points_per_curve = points_per_curve
        # Gate parameters
        self._gate_start = gate_start
        self._gate_width_1 = gate_width_1
        self._gate_width_2 = gate_width_2
        self._gate_length = gate_length
        self._gate_point_2 = gate_point_2
        self._gate_point_3 = gate_point_3

        # Choke parameters
        self._choke_start = choke_start
        self._choke_width_1 = choke_width_1
        self._choke_width_2 = choke_width_2
        self._gate_choke_length = gate_choke_length
        self._choke_length_2 = choke_length_2
        self._choke_length_3 = choke_length_3
        self._choke_point_2 = choke_point_2
        self._choke_point_3 = choke_point_3

        # Channel parameter
        self._channel_length = channel_length
        self._outer_channel_width = outer_channel_width
        self._inner_channel_width = inner_channel_width
        self._channel_point_2 = channel_point_2
        self._channel_point_3 = channel_point_3
        self._channel_position = channel_position

    # symmetric bezier shaped polygon
    def _bezier_guide_symm(self, start, width_1, width_2, length, p2, p3):
        top_curve = CubicBezierCurve(
            *[(start, width_1 / 2), (start + p2[0] * length, width_1 / 2 - p2[1] * width_2 / 2),
              (start + p3[0] * length, width_1 / 2 - p3[1] * (width_1 / 2 - width_2 / 2)),
              (start + length, width_2 / 2)])

        bottom_curve = CubicBezierCurve(
            *[(start, -width_1 / 2), (start + p2[0] * length, -width_1 / 2 + p2[1] * width_2 / 2),
              (start + p3[0] * length, -width_1 / 2 + p3[1] * (width_1 / 2 - width_2 / 2)),
              (start + length, -width_2 / 2)])

        y_bottom = bottom_curve.evaluate(np.linspace(1, 0, self._points_per_curve))
        y_top = top_curve.evaluate(np.linspace(0, 1, self._points_per_curve))

        return Polygon(zip(y_top[0], y_top[1]) + zip(y_bottom[0], y_bottom[1]))

    # one side flat and one side bezier shaped polygon
    def _bezier_guide_channel(self, start, outer_width, inner_width, length, p2, p3):
        top_curve = CubicBezierCurve(
            (start, outer_width), (start + p2[0] * length / 2., outer_width - p2[1] * inner_width),
            (start + p3[0] * length / 2., inner_width + p3[1] * (outer_width - inner_width)),
            (start + length / 2., inner_width))
        bottom_curve = CubicBezierCurve((start + length / 2., inner_width),
                                        (start + length / 2. + p2[0] * length / 2., inner_width + p2[1] * inner_width),
                                        (start + length / 2. + p3[0] * length / 2.,
                                         outer_width - p3[1] * (outer_width - inner_width)),
                                        (start + length, outer_width))

        return Polygon(zip(*top_curve.evaluate(np.linspace(0, 1, self._points_per_curve))) + zip(*bottom_curve.evaluate(
            np.linspace(0, 1, self._points_per_curve))) + ([(length, 0), (0, 0)]))

    # Creating the different par polygons
    def _gate(self):
        return self._bezier_guide_symm(self._gate_start, self._gate_width_1, self._gate_width_2, self._gate_length,
                                       self._gate_point_2, self._gate_point_3)

    def _choke_channel(self):
        return self._bezier_guide_symm(self._choke_start, self._choke_width_1, self._choke_width_1,
                                       self._gate_choke_length, self._gate_point_2, self._gate_point_2)

    def _choke_left(self):
        return self._bezier_guide_symm(self._choke_start + self._gate_choke_length, self._choke_width_1,
                                       self._choke_width_2, self._choke_length_2, self._gate_point_2,
                                       self._gate_point_3)

    def _choke_right(self):
        return self._bezier_guide_symm(self._gate_length + self._gate_choke_length + self._choke_length_2,
                                       self._choke_width_2, self._choke_width_1, self._choke_length_3,
                                       self._choke_point_2,
                                       self._choke_point_3)

    def _channel(self):
        channel_object = self._bezier_guide_channel(0, self._outer_channel_width, self._inner_channel_width,
                                                    self._channel_length, self._channel_point_2, self._channel_point_3)
        x_offset = (self._gate_length + self._gate_choke_length + self._choke_length_2 + self._choke_length_3)
        y_offset = (1 - self._channel_position) * self._channel_length
        return translate(rotate(channel_object, -np.pi / 2, (0, 0), True), x_offset, y_offset, 0)

    # Unite and translate the polygons parts
    def get_shapely_object(self):
        object = geometric_union(
            [self._gate(), self._choke_channel(), self._choke_left(), self._choke_right(), self._channel()])
        rotated_object = rotate(object, self._angle, (0, 0), True)
        return translate(rotated_object, self._origin[0], self._origin[1], 0)

    @classmethod
    def make_at_port_(cls, port, gate_width_1, gate_width_2, choke_width_1, choke_width_2, choke_length_2,
                      choke_length_3, choke_point_2=(1.0 / 2, 0), choke_point_3=(1, 1.0 / 2), points_per_curve=1000,
                      gate_start=0, gate_point_2=(1.0 / 3, 0), gate_point_3=(2.0 / 3, 1), channel_point_2=(1.0 / 3, 0),
                      channel_point_3=(2.0 / 3, 0), channel_length=2.5, outer_channel_width=0.4,
                      inner_channel_width=0.2,
                      channel_position=0.3, gate_length=0.5, choke_start=0.5, gate_choke_length=0.2, target='_gate'):
        """
        Class method to directly place a ntron to a given port, connected to a chosen port

        :param port: port from the connecting structure
        :param gate_width_1: outer width of the _gate port
        :param gate_width_2: connection gate-choke, should be same as choke_width_1
        :param choke_width_1: outer width of the choke
        :param choke_width_2: Inner choke width, which is the bottleneck and the most important parameter
        :param choke_length_2: length of the gate sided choke bezier shape
        :param choke_length_3: length of the channel sided choke bezier shape
        :param choke_point_2: 2nd bezier point in a rectangle (Look Bezier explanation above)
        :param choke_point_3: 3rd bezier point in a rectangle (Look Bezier explanation above)
        :param points_per_curve: number of of points per Bezier curve used to create a shapely geometry
        :param gate_start: starting x pos of the gate section, should be zero
        :param gate_point_2: 2nd bezier point in a rectangle (Look Bezier explanation above)
        :param gate_point_3: 3rd bezier point in a rectangle (Look Bezier explanation above)
        :param channel_point_2: 2nd bezier point in a rectangle (Look Bezier explanation above)
        :param channel_point_3: 3rd bezier point in a rectangle (Look Bezier explanation above)
        :param channel_length: length of the channel
        :param outer_channel_width: starting width of the channel at the gates
        :param inner_channel_width: end width in the middle of the channel
        :param channel_position: channel position corresponding to the gate, for 0 choke is on the bottom, for 1 on the top
        :param gate_length: the length of the bezier calculated shape of the gate
        :param choke_start: starting x pos of the choke section, should be equal to gate_length
        :param gate_choke_length:
        :param target: the ntron port to connect to
        :return:
        """

        # Calculation of the target relative parameters
        if target == 'gate':
            return cls(port.origin, port.angle, gate_width_1, gate_width_2, choke_width_1, choke_width_2,
                       choke_length_2,
                       choke_length_3, choke_point_2, choke_point_3, points_per_curve, gate_start, gate_point_2,
                       gate_point_3, channel_point_2, channel_point_3, channel_length, outer_channel_width,
                       inner_channel_width, channel_position, gate_length, choke_start, gate_choke_length)
        if target == 'source':
            x = (channel_position * channel_length)
            y = (gate_length + gate_choke_length + choke_length_2 + choke_length_3 + outer_channel_width / 2)

            x_off = port.origin[0] + np.cos(port.angle) * x - np.sin(port.angle) * y
            y_off = port.origin[1] + np.sin(port.angle) * x + np.cos(port.angle) * y

            return cls(
                (x_off, y_off), port.angle - np.pi / 2, gate_width_1, gate_width_2, choke_width_1, choke_width_2,
                choke_length_2,
                choke_length_3)
        if target == 'drain':
            x = (1 - channel_position) * channel_length
            y = -gate_length - gate_choke_length - choke_length_2 - choke_length_3 - outer_channel_width / 2

            x_off = port.origin[0] + np.cos(port.angle) * x - np.sin(port.angle) * y
            y_off = port.origin[1] + np.sin(port.angle) * x + np.cos(port.angle) * y

            return cls((x_off, y_off), port.angle + np.pi / 2, gate_width_1, gate_width_2, choke_width_1, choke_width_2,
                       choke_length_2, choke_length_3)

    @property
    def origin(self):
        return self._origin

    @property
    def angle(self):
        return self._angle

    @property
    def width(self):
        return self._gate_width_1

    @property
    def port_drain(self):
        x = (self._gate_length + self._gate_choke_length + self._choke_length_2 +
             self._choke_length_3 + self._outer_channel_width / 2.)
        y = ((1 - self._channel_position) * self._channel_length)

        x_off = self.origin[0] + np.cos(self._angle) * x - np.sin(self._angle) * y
        y_off = self.origin[1] + np.sin(self._angle) * x + np.cos(self._angle) * y

        return Port((x_off, y_off), self._angle + np.pi / 2., self._outer_channel_width)

    @property
    def port_gate(self):
        return Port(self.origin, self._angle + np.pi, self._gate_width_1)

    @property
    def port_source(self):
        x = (self._gate_length + self._gate_choke_length + self._choke_length_2 + self._choke_length_3
             + self._outer_channel_width / 2.)
        y = (-self._channel_position * self._channel_length)

        x_off = self.origin[0] + np.cos(self._angle) * x - np.sin(self._angle) * y
        y_off = self.origin[1] + np.sin(self._angle) * x + np.cos(self._angle) * y

        return Port((x_off, y_off), self._angle - np.pi / 2, self._outer_channel_width)


if __name__ == '__main__':
    import gdsCAD.core

    part1 = Ntron(origin=(5, -5), angle=0, gate_width_1=0.3, gate_width_2=0.06, choke_width_1=0.06, choke_width_2=0.015,
                  choke_length_2=0.06, choke_length_3=0.03)
    wgdrain = Waveguide.make_at_port(part1.port_drain)
    wgdrain.add_straight_segment(0.5)
    wggate = Waveguide.make_at_port(part1.port_gate)
    wggate.add_straight_segment(0.5)
    wgsource = Waveguide.make_at_port(part1.port_source)
    wgsource.add_straight_segment(0.5)
    cell = gdsCAD.core.Cell('_channel')
    cell.add(convert_to_gdscad([part1], layer=1))
    cell.show()



import numpy as np


class Alignment(object):
    """
    Alignment helper class.

    Given a bounding box in the form of ``((min_x, min_y), (max_x, max_y))``, this class
    returns special points of this box::

        (max_y)    top +------+------+
                     |      |      |
                     |      |      |
             center  +------+------+
                     |      |      |
                     |      |      |
        (min_y) bottom +------+------+
                  (min_x)         (max_x)
                    Left  Center   Right

    Alignment options are given as ``-`` separated tuple, allowing for combinations of
    ``left``, ``center``, ``right`` with ``bottom``, ``center``, ``top``.
    """
    _ALIGNMENT = {
        'x': {
            'left': lambda coord: coord[0][0],
            'center': lambda coord: np.mean(coord[:, 0]),
            'right': lambda coord: coord[1][0],
        },
        'y': {
            'bottom': lambda coord: coord[0][1],
            'center': lambda coord: np.mean(coord[:, 1]),
            'top': lambda coord: coord[1][1],
        }
    }

    def __init__(self, alignment='bottom-left'):
        self.alignment = alignment

    @property
    def alignment(self):
        """
        Property holding the current alignment.

        :return: Alignment string, i.e. ``'bottom-left'``.
        :rtype: str
        """
        return self._alignment

    # noinspection PyAttributeOutsideInit
    @alignment.setter
    def alignment(self, alignment):
        assert type(alignment) == str, 'Alignment must be a string'

        options = [option.strip() for option in alignment.split('-')]
        assert len(options) == 2, 'Alignment option string must be two options separated by a dash'
        assert options[0] in self._ALIGNMENT['x'], 'x-axis alignment option must be one of %s' % \
                                                   self._ALIGNMENT['x'].keys()
        assert options[1] in self._ALIGNMENT['y'], 'y-axis alignment option must be one of %s' % \
                                                   self._ALIGNMENT['y'].keys()

        self._alignment = '-'.join(options)

    @property
    def alignment_functions(self):
        """
        Returns a 2-tuple of functions, calculating the offset coordinates for a given bounding box.

        :return: Tuple of functions.
        :rtype: tuple
        """
        options = self.alignment.split('-')
        return self._ALIGNMENT['x'][options[0]], self._ALIGNMENT['y'][options[1]]

    def calculate_offset(self, bbox):
        """
        Calculate the coordinates of the current alignment for the given bounding box *bbox*.
        
        :param bbox: Bounding box in the ``((min_x, min_y), (max_x, max_y))`` format.
        :return: (x, y) offset coordinates.
        :rtype: np.array
        """
        bbox = np.asarray(bbox)
        alignment_fun = self.alignment_functions
        offset = np.array([-alignment_fun[i](bbox) for i in [0, 1]])
        return offset


import numpy as np
import shapely.geometry
import shapely.affinity

from gdshelpers.geometry import geometric_union


class SquareMarker(object):
    """ Square Marker

        :param origin: (x,y), center of square
        :type origin: tuple
        :param size: length
        :type size: float or int
        """

    def __init__(self, origin, size):
        assert (size >= 0)
        self.origin = np.asarray(origin)
        self.size = size

    @classmethod
    def make_marker(cls, origin, size=20):
        return cls(origin, size)

    def get_shapely_object(self):
        x, y = self.origin
        size = self.size
        box = shapely.geometry.box(0, 0, size, size)
        return shapely.affinity.translate(box, x - size / 2, y - size / 2)


class CrossMarker(object):
    """
    Simple cross type marker with support for paddles.

    This code is an adapted version of Nico's marker.
    """

    def __init__(self, origin, cross_length, cross_width, paddle_length, paddle_width):
        assert (cross_length >= 0) and (cross_width >= 0) and (paddle_length >= 0) and (paddle_width >= 0)

        self.origin = np.asarray(origin)
        self.paddle_length = paddle_length
        self.paddle_width = paddle_width
        self.cross_length = cross_length
        self.cross_width = cross_width

    @classmethod
    def make_simple_cross(cls, origin, cross_length, cross_width):
        return cls(origin, cross_length, cross_width, 0., 0.)

    @classmethod
    def make_traditional_paddle_markers(cls, origin, scale=1.,
                                        cross_length_factor=2.0, cross_width_factor=0.1,
                                        paddle_length_factor=3.0, paddle_width_factor=0.5):
        wp = paddle_width_factor * scale
        wc = cross_width_factor * scale
        lp = paddle_length_factor * scale
        lc = cross_length_factor * scale
        return cls(origin, lc, wc, lp, wp)

    def get_shapely_object(self):
        x, y = self.origin
        lp, wp, lc, wc = self.paddle_length, self.paddle_width, self.cross_length, self.cross_width

        points = [(x + wc, y + wc), (x + wc, y + lc), (x + wp, y + lc),
                  (x + wp, y + lc + lp), (x - wp, y + lc + lp), (x - wp, y + lc),
                  (x - wc, y + lc), (x - wc, y + wc), (x - lc, y + wc),
                  (x - lc, y + wp), (x - lc - lp, y + wp), (x - lc - lp, y - wp),
                  (x - lc, y - wp), (x - lc, y - wc), (x - wc, y - wc), (x - wc, y - lc),
                  (x - wp, y - lc), (x - wp, y - lc - lp), (x + wp, y - lc - lp), (x + wp, y - lc),
                  (x + wc, y - lc), (x + wc, y - wc), (x + lc, y - wc), (x + lc, y - wp),
                  (x + lc + lp, y - wp), (x + lc + lp, y + wp), (x + lc, y + wp), (x + lc, y + wc)]

        return shapely.geometry.Polygon(points)


class DLWMarker(object):
    def __init__(self, origin, box_size=2.5):
        self.origin = np.asarray(origin)
        self.box_size = box_size

    def get_shapely_object(self):
        x, y = self.origin
        upper_box = shapely.geometry.box(x, y, x + self.box_size, y + self.box_size)
        lower_box = shapely.geometry.box(x - self.box_size, y - self.box_size, x, y)
        return shapely.geometry.MultiPolygon([upper_box, lower_box])


class DLWPrecisionMarker(object):
    """
    A specific marker to test the writing accuracy/alignment of the Nanoscribe DLW machine.

    :param origin: Position of the marker.
    :param size: Size of the marker.
    :param frame_width: Size of the marker frame.
    """

    def __init__(self, origin, size, frame_width):
        self.origin = origin
        self.size = size
        self.frame_width = frame_width

    def get_shapely_object(self):
        size_half = self.size / 2.
        p1 = shapely.geometry.Polygon(
            [(-size_half, size_half + self.frame_width), (size_half, size_half + self.frame_width),
             (0, size_half)])
        p2 = shapely.geometry.Polygon([(size_half, size_half), (size_half, size_half + self.frame_width),
                                       (size_half + self.frame_width, size_half)])

        one_side = geometric_union([p1, p2])

        marker = geometric_union([shapely.affinity.rotate(one_side, phi, (0, 0)) for phi in (0, 90, 180, 270)])
        return shapely.affinity.translate(marker, self.origin[0], self.origin[1])


class AutoStigmationMarker(object):
    def __init__(self, origin, maximum_feature_size=3., minimum_feature_size=0.1, reduction_factor=1.2, resolution=16):
        self.origin = origin
        self.maximum_feature_size = maximum_feature_size
        self.minimum_feature_size = minimum_feature_size
        self.reduction_factor = reduction_factor
        self.resolution = resolution

    def get_shapely_object(self):
        radius = self.minimum_feature_size/2
        feature_size = self.minimum_feature_size

        center_point = shapely.geometry.Point(self.origin[0], self.origin[1])
        objs = list()

        # The inner object is a circle
        objs.append(center_point.buffer(self.minimum_feature_size/2., resolution=self.resolution))

        while feature_size < self.maximum_feature_size:
            feature_size *= self.reduction_factor
            radius += feature_size*2

            ring = center_point.buffer(radius+feature_size, self.resolution).difference(
                center_point.buffer(radius, self.resolution))
            objs.append(ring)

        return shapely.geometry.MultiPolygon(objs)


#: Layer for photonic devices (Waveguides, Couplers, Splitters...)
outlayer = 1

#: Layer for global markers (Crosses, ...)
gmarklayer = 2

#: Layer for local markers (Crosses, ...)
lmarklayer = 3

#: Layer for write fields. (Used just for checking, not used in lithography)
wflayer = 4

#: Layer for region definition. (For defining working areas = multiple of writefield_size)
regionlayer = 5

#: Layer for frames. (For overall pattern and/or arrays of devices)
framelayer = 6

#: Layer for pattern names. (For example NG01 CalibCoupler)
patnamelayer = 7

#: Layer for device names. (For example A1, F0, etc..)
devnamelayer = 8

#: Layer for local parameters. (Period of couplers, gaps, ...  Used for first lithography step)
parnamelayer1 = 9

#: Layer for local parameters. (Period of couplers, gaps, ...  Used for second lithography step)
parnamelayer2 = 10

#: Metal contact pad layer. (For probe contacts or wire bonding)
padlayer = 11

#: Wing layer. (For connection between nanowires and pads)
winglayer = 12

#: Nano features layer. (SSPD wires, high resolution elements...)
nanolayer = 13

#: Mask window layer 1. (opening windows for subsequent etching e.g. freestanding devices)
masklayer1 = 14

#: Mask window layer 2. (opening windows for subsequent etching e.g. freestanding devices)
masklayer2 = 15

#!: List of unassigned layers for general purpose use.
gplayers = tuple(range(16, 70))


import itertools
import json
import numpy as np
from collections import defaultdict
from shapely.affinity import translate, rotate
from shapely.geometry import box

from gdshelpers.geometry.shapely_adapter import convert_to_layout_objs, bounds_union, transform_bounds
from gdshelpers.parts.waveguide import Waveguide
from gdshelpers.parts.marker import DLWMarker
from gdshelpers.parts.text import Text
from gdshelpers.geometry import geometric_union
import gdshelpers.helpers.layers as std_layers

automatic_library = None
try:
    import gdspy

    automatic_library = 'gdspy'
except ImportError:
    pass  # Error will come when trying to use get_gdspy_cell
try:
    import gdsCAD

    automatic_library = 'gdscad'
except ImportError:
    pass  # Error will come when trying to use get_gdscad_cell

try:
    import fatamorgana
    import fatamorgana.records
except ImportError:
    pass


class Cell:
    def __init__(self, name):
        """
        Creates a new Cell named `name` at `origin`

        :param name: Name of the cell, needs to be unique
        """
        self.name = name
        self.cells = []
        self.layer_dict = defaultdict(lambda: [])
        self.dlw_data = {}
        self.desc = {'dlw': self.dlw_data, 'desc': {}, 'ebl': []}
        self.cell_gdspy = None
        self.cell_gdscad = None
        self.cell_oasis = None
        self._bounds = None
        # Only contains the bounds of items in `layer_dict`.
        # Children cells have to be queried each time, since there is no way of knowing when they got changed.
        # self._bounds is None if the bounds need to be recalculated or if the cell is empty (in which case
        # recalculating them is cheap and we don't need to cache them)

    @property
    def bounds(self):
        """
        The outer bounding box of the cell. Returns `None` if it is empty.
        """
        return self.get_bounds(layers=None)

    def get_bounds(self, layers=None):
        """
        Calculates and returns the envelope for the given layers.
        Returns `None` if it is empty.
        """
        # Go through all layers
        bounds = []
        if layers is None and self._bounds is not None:
            bounds += [self._bounds]
        else:
            for layer in layers or self.layer_dict.keys():
                for geo in self.layer_dict.get(layer, []):
                    geo_bounds = geo.get_shapely_object().bounds if hasattr(geo, 'get_shapely_object') else geo.bounds
                    if geo_bounds is not ():  # Some shapely geometries (collections) can return empty bounds
                        bounds.append(geo_bounds)

            # Cache envelope if we have the global envelope
            if layers is None:
                self._bounds = bounds_union(bounds) if len(bounds) > 0 else None

        # Merge envelopes of children cells
        for cell in self.cells:
            cell_bounds = cell['cell'].get_bounds(layers)
            if cell_bounds is not None:
                bounds.append(transform_bounds(cell_bounds, cell['origin'], rotation=cell['angle'] or 0))

        return bounds_union(bounds) if len(bounds) > 0 else None

    @property
    def size(self):
        bounds = self.bounds
        if bounds is None:
            return 0, 0
        else:
            return bounds[2] - bounds[0], bounds[3] - bounds[1]

    def add_to_layer(self, layer, *geometry):
        """
        Adds a shapely geometry to a the layer

        :param layer: id of the layer
        :param geometry: shapely geometry
        """
        self._bounds = None
        self.layer_dict[layer] += geometry

    def add_dlw_data(self, dlw_type, dlw_id, data):
        """
        Adds data for 3D-hybrid-integration to the Cell. This is usually only done by using the Device

        :param dlw_type: type of the represented object
        :param dlw_id: id of the represented object
        :param data: data of the object
        """
        if dlw_type not in self.dlw_data:
            self.dlw_data[dlw_type] = {}
        if dlw_id in self.dlw_data[dlw_type]:
            raise ValueError('ID "{:s}" already used'.format(dlw_id))
        self.dlw_data[dlw_type][dlw_id] = data

    def add_cell(self, cell, origin=(0, 0), angle=None):
        """
        Adds a Cell to this cell

        :param cell: Cell to add
        :param origin: position where to add the cell
        :param angle:
        """
        if cell.name in [cell['cell'].name for cell in self.cells]:
            import warnings
            warnings.warn(
                'Cell name "{cell_name:s}" added multiple times to {self_name:s}.'
                ' Can be problematic for desc/dlw-files'.format(cell_name=cell.name, self_name=self.name))
        self.cells.append(dict(cell=cell, origin=origin, angle=angle))

    def add_region_layer(self, region_layer=std_layers.regionlayer, layers=None):
        """
        Generate a region layer around all objects on `layers` and place it on layer `region_layer`.
        If `layers` is None, all layers are used.
        """
        self.add_to_layer(region_layer, box(*self.get_bounds(layers)))

    def add_frame(self, padding=30., line_width=1., frame_layer=std_layers.framelayer, bbox=None):
        """
        Generates a rectangular frame around the contents of the cell.

        :param padding: Add a padding of the given value around the contents of the cell
        :param line_width: Width of the frame line
        :param frame_layer: Layer to put the frame on.
        :param bbox: Optionally, an explicit extent can be passed to the function. If `None` (default),
                     the current extent of the cell will be chosen.
        """
        padding = padding + line_width
        bbox = bbox or self.bounds

        frame = box(bbox[0] - padding, bbox[1] - padding, bbox[2] + padding, bbox[3] + padding)
        frame = frame.difference(frame.buffer(-line_width))
        self.add_to_layer(frame_layer, frame)

    def add_ebl_marker(self, layer, marker):
        """
        Adds an Marker to the layout

        :param layer: layer on which the marker should be positioned
        :param marker: marker, that should be added (from gdshelpers.parts.markers)
        """
        self.add_to_layer(layer, marker)
        self.desc['ebl'].append(list(marker.origin))

    def add_ebl_frame(self, layer, frame_generator, **kwargs):
        """
        Adds global markers to the layout

        :param layer:  layer on which the markers should be positioned
        :param frame_generator: either a method, which returns a list of the markers, which should be added or the name
            of a generator from the gdshelpers.geometry.ebl_frame_generators package
        :param kwargs: Parameters which are directly passed to the frame generator
        """
        from gdshelpers.geometry import ebl_frame_generators
        frame_generator = frame_generator if callable(frame_generator) else getattr(ebl_frame_generators,
                                                                                    frame_generator)
        for marker in frame_generator(self.bounds, **kwargs):
            self.add_ebl_marker(layer, marker)

    def add_to_desc(self, key, data):
        """
        Adds data to the .desc-file can be used for any data related to the cells, e.g. the swept parameters/...

        :param key: name of the entry
        :param data: data which are added to the .desc-file, the data have to be serializable by json
        """
        self.desc['desc'][key] = data

    def get_dlw_data(self):
        dlw_data = self.dlw_data.copy()
        for sub_cell in self.cells:
            cell, origin = sub_cell['cell'], sub_cell['origin']

            for dlw_type, dlw_type_data in cell.get_dlw_data().items():
                for dlw_id, data in dlw_type_data.items():
                    data = data.copy()
                    data['origin'] = (np.array(origin) + data['origin']).tolist()
                    if dlw_type not in dlw_data:
                        dlw_data[dlw_type] = {}
                    dlw_data[dlw_type][cell.name + '.' + dlw_id] = data

        return dlw_data

    def get_desc(self):
        desc = self.desc.copy()
        desc['cells'] = {cell['cell'].name: dict(offset=tuple(cell['origin']), angle=cell['angle'] or 0,
                                                 **cell['cell'].get_desc()) for cell in self.cells}
        return desc

    def get_gdspy_cell(self, executor=None):
        if self.cell_gdspy is None:
            self.cell_gdspy = gdspy.Cell(self.name)
            for sub_cell in self.cells:
                angle = np.rad2deg(sub_cell['angle']) if sub_cell['angle'] is not None else None
                self.cell_gdspy.add(
                    gdspy.CellReference(sub_cell['cell'].get_gdspy_cell(executor), origin=sub_cell['origin'],
                                        rotation=angle))
            for layer, geometry in self.layer_dict.items():
                if executor:
                    executor.submit(convert_to_layout_objs, geometry, layer, library='gdspy') \
                        .add_done_callback(lambda future: self.cell_gdspy.add(future.result()))
                else:
                    self.cell_gdspy.add(convert_to_layout_objs(geometry, layer, library='gdspy'))
        return self.cell_gdspy

    def get_gdscad_cell(self, executor=None):
        if self.cell_gdscad is None:
            self.cell_gdscad = gdsCAD.core.Cell(self.name)
            for sub_cell in self.cells:
                angle = np.rad2deg(sub_cell['angle']) if sub_cell['angle'] is not None else None
                self.cell_gdscad.add(
                    gdsCAD.core.CellReference(sub_cell['cell'].get_gdscad_cell(executor), origin=sub_cell['origin'],
                                              rotation=angle))
            for layer, geometry in self.layer_dict.items():
                if executor:
                    executor.submit(convert_to_layout_objs, geometry, layer, library='gdscad') \
                        .add_done_callback(lambda future: self.cell_gdscad.add(future.result()))
                else:
                    self.cell_gdscad.add(convert_to_layout_objs(geometry, layer, library='gdscad'))
        return self.cell_gdscad

    def get_oasis_cells(self, grid_steps_per_micron=1000, executor=None):
        if self.cell_oasis is None:
            self.cell_oasis = fatamorgana.Cell(fatamorgana.NString(self.name))
            for sub_cell in self.cells:
                x, y = sub_cell['origin']
                x, y = round(x * grid_steps_per_micron), round(y * grid_steps_per_micron)
                angle = np.rad2deg(sub_cell['angle']) if sub_cell['angle'] is not None else None
                self.cell_oasis.placements.append(
                    fatamorgana.records.Placement(False, name=fatamorgana.NString(sub_cell['cell'].name), x=x, y=y,
                                                  angle=angle))
            for layer, geometry in self.layer_dict.items():
                if executor:
                    executor.submit(convert_to_layout_objs, geometry, layer, library='oasis',
                                    grid_steps_per_micron=grid_steps_per_micron, max_points=np.inf,
                                    max_points_line=np.inf) \
                        .add_done_callback(lambda future: self.cell_oasis.geometry.extend(future.result()))
                else:
                    self.cell_oasis.geometry.extend(convert_to_layout_objs(geometry, layer, library='oasis',
                                                                           grid_steps_per_micron=grid_steps_per_micron,
                                                                           max_points=np.inf, max_points_line=np.inf))
        return [self.cell_oasis] + [oasis_cell for sub_cell in self.cells for oasis_cell in
                                    sub_cell['cell'].get_oasis_cells(grid_steps_per_micron, executor)]

    def get_gdspy_lib(self):
        self.get_gdspy_cell()
        return gdspy.current_library

    def start_viewer(self):
        gdspy.LayoutViewer(library=self.get_gdspy_lib(), depth=10)

    def save(self, name=None, library=automatic_library, grid_steps_per_micron=1000, parallel=False):
        """
        Exports the layout and creates an DLW-file, if DLW-features are used.

        :param name: Optionally, the filename of the saved file (without ending).
        :param library: Name of the used library.
            Currently, for gds-export gdspy and gdscad are supported, for oasis-export fatamorgana is supported.
        :param grid_steps_per_micron: Defines the resolution
        :param parallel: Defines if palatalization is used (only supported in Python 3).
            Standard value will be changed to True in a future version.
            Deactivating can be useful for debugging reasons.
        """

        if not name:
            name = self.name

        if library == 'gdspy':
            if parallel:
                from concurrent.futures import ProcessPoolExecutor
                with ProcessPoolExecutor() as pool:
                    self.get_gdspy_cell(pool)
            else:
                self.get_gdspy_cell()

            self.get_gdspy_lib().precision = self.get_gdspy_lib().unit / grid_steps_per_micron
            gdspy_cells = self.get_gdspy_lib().cell_dict.values()
            if parallel:
                from concurrent.futures import ProcessPoolExecutor
                with ProcessPoolExecutor() as pool:
                    binary_cells = pool.map(gdspy.Cell.to_gds, gdspy_cells, [grid_steps_per_micron] * len(gdspy_cells))
            else:
                binary_cells = map(gdspy.Cell.to_gds, gdspy_cells, [grid_steps_per_micron] * len(gdspy_cells))

            self.get_gdspy_lib().write_gds(name + '.gds', cells=[], binary_cells=binary_cells)
        elif library == 'gdscad':
            layout = gdsCAD.core.Layout(precision=1e-6 / grid_steps_per_micron)
            if parallel:
                from concurrent.futures import ProcessPoolExecutor
                with ProcessPoolExecutor() as pool:
                    layout.add(self.get_gdscad_cell(pool))
            else:
                layout.add(self.get_gdscad_cell())
            layout.save(name + '.gds')
        elif library == 'fatamorgana':
            layout = fatamorgana.OasisLayout(grid_steps_per_micron)

            if parallel:
                from concurrent.futures import ProcessPoolExecutor
                with ProcessPoolExecutor() as pool:
                    cells = self.get_oasis_cells(grid_steps_per_micron, pool)
            else:
                cells = self.get_oasis_cells(grid_steps_per_micron)

            layout.cells = [cells[0]] + list(set(cells[1:]))

            name_id = {}
            for i, cell in enumerate(layout.cells):
                if cell.name.string in name_id:
                    raise RuntimeError(
                        'Each cell name should be unique, name "' + cell.name.string + '" is used multiple times')
                name_id[cell.name.string] = i
                cell.name = i
            for cell in layout.cells:
                for placement in cell.placements:
                    placement.name = name_id[placement.name.string]

            layout.cellnames = {v: k for k, v in name_id.items()}

            with open(name + '.oas', 'wb') as f:
                layout.write(f)
        else:
            raise ValueError('library must be either "gdscad", "gdspy" or "fatamorgana"')

        dlw_data = self.get_dlw_data()
        if dlw_data:
            with open(name + '.dlw', 'w') as f:
                json.dump(dlw_data, f, indent=True)

        with open(name + '.desc', 'w') as f:
            json.dump(self.get_desc(), f, indent=True)

    def get_reduced_layer(self, layer):
        def translate_and_rotate(geometry, offset, angle):
            if not geometry:
                return geometry
            return translate(rotate(geometry, angle if angle else 0, use_radians=True, origin=(0, 0)), *offset)

        return geometric_union(
            (self.layer_dict[layer] if layer in self.layer_dict else []) +
            [translate_and_rotate(cell['cell'].get_reduced_layer(layer), cell['origin'], cell['angle'])
             for cell in self.cells])

    def export_mesh(self, filename, layer_defs):
        from functools import reduce
        from trimesh.primitives import Extrusion
        from trimesh.transformations import translation_matrix

        reduce(lambda a, b: a + b, (Extrusion(polygon=geometry, height=min_max[1] - min_max[0],
                                              transform=translation_matrix((0, 0, min_max[0])))
                                    for layer, min_max in layer_defs.items()
                                    for geometry in self.get_reduced_layer(layer))).export(filename)

    def _get_patches(self, origin=(0, 0), angle_sum=0, angle=0, layers=None):
        from descartes import PolygonPatch

        def rotate_pos(pos, angle):
            if angle is None:
                return pos
            c, s = np.cos(angle), np.sin(angle)
            result = np.array([[c, -s], [s, c]]).dot(pos)
            return result

        own_patches = [
            PolygonPatch(
                translate(rotate(geometric_union(geometry), angle_sum, use_radians=True, origin=(0, 0)), *origin), color=['red', 'green', 'blue', 'teal', 'pink'][(layer-1)%5], linewidth=0)
            for layer, geometry in self.layer_dict.items() if (layers is None or layer in layers)]
        sub_cells_patches = [p for cell_dict in self.cells for p in
                             cell_dict['cell']._get_patches(
                                 np.array(origin) + rotate_pos(cell_dict['origin'], angle),
                                 angle_sum=angle_sum + (cell_dict['angle'] or 0), angle=cell_dict['angle'],
                                 layers=layers)]

        return own_patches + sub_cells_patches

    def show(self, layers=None, padding=5):
        """
        Shows the current cell

        :param layers: List of the layers to be shown, passing None shows all layers
        :param padding: padding around the structure
        """
        import matplotlib.pyplot as plt

        fig, ax = plt.subplots()
        for patch in self._get_patches(layers=layers):
            ax.add_patch(patch)

        bounds = self.get_bounds(layers)
        ax.set_xlim(bounds[0] - padding, bounds[2] + padding)
        ax.set_ylim(bounds[1] - padding, bounds[3] + padding)
        ax.set_aspect(1)
        fig.show()

    def add_dlw_marker(self, label, layer, origin):
        """
        Adds a marker for 3D-hybrid integration

        :param label: Name of the marker, needs to be unique within the device
        :param layer: Layer at which the marker and markers should be written
        :param origin: Position of the marker
        """
        self.add_to_layer(layer, DLWMarker(origin))
        self.add_to_layer(std_layers.parnamelayer1, Text(origin, 2, label, alignment='center-center'))

        self.add_dlw_data('marker', label, {'origin': origin.tolist()})

    def add_dlw_taper_at_port(self, label, layer, port, taper_length, tip_width=.01, with_markers=True):
        """
        Adds a taper for 3D-hybrid-integration at a certain port

        :param label: Name of the port, needs to be unique within the device
        :param layer: Layer at which the taper and markers should be written
        :param port: Port to which the taper should be attached
        :param taper_length: length of the taper
        :param tip_width: final width of the tip
        :param with_markers: for recognizing the taper markers near to the taper are necessary.
            In certain designs the standard positions are not appropriate and can therefore be disabled and manually added
        """
        taper_port = port.longitudinal_offset(taper_length)
        if taper_length > 0:
            wg = Waveguide.make_at_port(port)
            wg.add_straight_segment(taper_length, final_width=tip_width)
            self.add_to_layer(layer, wg.get_shapely_object())
        self.add_to_layer(std_layers.parnamelayer1, Text(taper_port.origin, 2, label, alignment='center-center'))

        self.add_dlw_data('taper', str(label), {'origin': taper_port.origin.tolist(), 'angle': port.angle,
                                                'starting_width': port.width, 'taper_length': taper_length})
        if with_markers:
            for i, (v, l) in enumerate(itertools.product((-20, 20), (taper_length, 0))):
                self.add_dlw_marker(str(label) + '-' + str(i), layer,
                                    port.parallel_offset(v).longitudinal_offset(l).origin)


if __name__ == '__main__':
    from gdshelpers.parts.port import Port
    from gdshelpers.parts.waveguide import Waveguide
    from gdshelpers.geometry.chip import Cell

    # Create a cell-like object that offers a save output command '.save' which creates the .gds or .oas file by using
    # gdsCAD,gdspy or fatamorgana
    device_cell = Cell('my_cell')
    # Create a port to connect waveguide structures to
    port = Port(origin=(0, 0), width=1, angle=0)
    waveguide = Waveguide.make_at_port(port)
    for i in range(9):
        waveguide.add_bend(angle=np.pi, radius=60 + i * 40)
    # Add direct laser writing taper and alignment marker for postprocessing with a dlw printer to the cell-like object.
    # The cell dlw files will be saved with the cell.
    device_cell.add_dlw_taper_at_port('A0', 2, port.inverted_direction, 30)
    device_cell.add_dlw_taper_at_port('A1', 2, waveguide.current_port, 30)
    device_cell.add_to_layer(1, waveguide)
    device_cell.show()
    # Creates the output file by using gdspy,gdsCAD or fatamorgana. To use the implemented parallell processing, set
    # parallel=True.
    device_cell.save(name='my_design', parallel=True, library='gdspy')



import numpy as np
from gdshelpers.geometry import geometric_union, shapely_adapter
from gdshelpers.parts.coupler import GratingCoupler


def create_holes_for_under_etching(underetch_parts, complete_structure, hole_radius, hole_distance, hole_spacing):
    """
    Creates holes around given parts which can be used for underetching processes

    :param underetch_parts: List of gdshelper parts around which the holes shall be placed
    :param complete_structure: geometric union of the complete structure, needed to avoid collisions between underetching holes and other structures, e.g. waveguides
    :param hole_radius: Radius of the holes in microns
    :param hole_distance: Distance of the holes center from the the structures in microns
    :param hole_spacing: Distance between the holes in microns
    :return: Geometric union of the created holes
    """

    union = geometric_union(underetch_parts)
    no_hole_zone = complete_structure.buffer(0.9 * hole_distance, resolution=32, cap_style=3)
    poly = union.buffer(hole_distance, resolution=32, cap_style=3)

    base_polygon = shapely_adapter.shapely_collection_to_basic_objs(poly)

    holes = []
    for obj in base_polygon:
        ext = obj.exterior
        for dist in np.arange(0, ext.length, hole_spacing):
            pos = ext.interpolate(distance=dist)
            if not no_hole_zone.contains(pos):
                holes.append(pos.buffer(hole_radius))
        for interior in obj.interiors:
            for dist in np.arange(0, interior.length, hole_spacing):
                pos = interior.interpolate(distance=dist)
                if not no_hole_zone.contains(pos):
                    holes.append(pos.buffer(hole_radius))

    return geometric_union(holes)


if __name__ == '__main__':
    import gdsCAD.core
    from gdshelpers.parts.waveguide import Waveguide
    from gdshelpers.parts.resonator import RingResonator
    from gdshelpers.geometry import convert_to_gdscad

    # ==== create some sample structures (straight line with ring resonator)
    wg = Waveguide(origin=(0, 0), angle=np.deg2rad(-90), width=1)
    wg.add_straight_segment(length=5)
    wg.add_bend(np.pi / 2, 5)
    wg2 = Waveguide.make_at_port(wg.current_port)
    wg2.add_straight_segment(15)
    reso = RingResonator.make_at_port(port=wg2.current_port, gap=0.2, radius=5)
    wg2.add_straight_segment(length=15)
    coupler2 = GratingCoupler.make_traditional_coupler_from_database_at_port(wg2.current_port, db_id='si220',
                                                                             wavelength=1550)

    underetching_parts = geometric_union([wg2, reso, coupler2])
    structure = geometric_union([underetching_parts, wg])
    # create the holes with a radius of 0.5 microns, a distance of 2 microns to the structure borders and
    # a distance of 2 microns between the holes
    holes = create_holes_for_under_etching(underetch_parts=underetching_parts, complete_structure=structure,
                                           hole_radius=0.5, hole_distance=2, hole_spacing=3)

    # create a cell with the structures in layer 1 and the holes in layer 2
    cell = gdsCAD.core.Cell('CELL')
    cell.add(convert_to_gdscad(structure, layer=1))
    cell.add(convert_to_gdscad(holes, layer=2))
    # Show the cell
    cell.show()



from itertools import product

from gdshelpers.parts.marker import SquareMarker


def raith_marker_frame(bounds, padding=100, distance=200, size=20):
    d1 = padding + distance / 2
    d2 = distance / 2

    return (SquareMarker((x + offset_x, y + offset_y), size)
            for x, y, offset_x, offset_y in
            product([bounds[0] - d1, bounds[2] + d1], [bounds[1] - d1, bounds[3] + d1], [-d2, d2], [-d2, d2]))



import numpy as np

import shapely.geometry
from gdshelpers.helpers import StandardLayers
from gdshelpers.geometry import convert_to_gdscad


def annotate_write_fields(cell, origin='left-top', end='right-bottom', size=100, layer=StandardLayers.wflayer):
    """
    Add write field marker lines to a cell.

    Write fields are rectangular and typically start at the upper left part of a structure. This function adds
    write field lines to a given cell. The bounds are either found automatically or can be given as (x, y) tuple.

    Alignment options are given as ``-`` separated tuple, allowing for combinations of ``left``, ``right``
    with ``bottom``, ``top``.

    For reasons of backward compatibility, the keyword strings ``upperleft``, ``upperright``, ``bottomleft``,
    ``bottomright`` are valid parameters for origin and end as well.


    :param cell: The cell in which the lines will be added.
    :param origin: Origin of the write field markers.
    :param end: End of the write field markes. Must not be aligned to the write field, however.
    :param size: Size of the write field squares.
    :param layer: Layer on which the write field lines are added.
    """
    corner_labels = {
        # Old alignment names
        'upperleft': lambda c: (c.bounding_box[0][0], c.bounding_box[1][1]),
        'upperright': lambda c: (c.bounding_box[1][0], c.bounding_box[1][1]),
        'bottomleft': lambda c: (c.bounding_box[0][0], c.bounding_box[0][1]),
        'bottomright': lambda c: (c.bounding_box[1][0], c.bounding_box[0][1]),

        # New alignment names
        'left-top': lambda c: (c.bounding_box[0][0], c.bounding_box[1][1]),
        'right-to': lambda c: (c.bounding_box[1][0], c.bounding_box[1][1]),
        'left-bottom': lambda c: (c.bounding_box[0][0], c.bounding_box[0][1]),
        'right-bottom': lambda c: (c.bounding_box[1][0], c.bounding_box[0][1])
    }

    assert type(origin) in [list, tuple] or origin in corner_labels.keys(), \
        'origin must either be a tuple or one of %s' % origin.keys()
    assert type(end) in [list, tuple] or end in corner_labels.keys(), \
        'origin must either be a tuple or one of %s' % corner_labels.keys()
    assert size > 0, 'Write field alignment must be a number greater 0'

    # Map string values to true coordinates
    origin = corner_labels[origin](cell) if origin in corner_labels else np.asarray(origin)
    end = corner_labels[end](cell) if end in corner_labels else float(end)

    # Calculate wf x and y positions
    x_begin = origin[0]
    # x_end = (end[0] // size) * (size + (1 if end[0] > origin[0] else -1))
    x_end = end[0]
    x_increasing = x_begin < x_end
    x_pos = np.arange(x_begin, x_end, size) if x_increasing else np.arange(x_begin, x_end, -size)[::-1]

    y_begin = origin[1]
    # y_end = (end[1] // size) * (size + (1 if end[1] > origin[1] else -1))
    y_end = end[1]
    y_increasing = y_begin < y_end
    y_pos = np.arange(y_begin, y_end, size) if y_increasing else np.arange(y_begin, y_end, -size)[::-1]

    x_size = size * (1 if x_increasing else -1)
    y_size = size * (1 if y_increasing else -1)
    for x in x_pos:
        for y in y_pos:
            cell.add(convert_to_gdscad(shapely.geometry.box(x, y, x + x_size, y + y_size), layer=layer))



import unittest
from gdshelpers.helpers import int_to_alphabet, id_to_alphanumeric


class HelpersTestCase(unittest.TestCase):
    def test_int_to_alphabet(self):
        self.assertEqual(int_to_alphabet(0), 'A')
        self.assertEqual(int_to_alphabet(25), 'Z')
        self.assertEqual(int_to_alphabet(26), 'AA')
        self.assertEqual(int_to_alphabet(26+25), 'AZ')

    def test_id_to_alphanumeric(self):
        self.assertEqual(id_to_alphanumeric(4, 26+25), 'AZ4')


def test_suite():
    return unittest.TestLoader().loadTestsFromTestCase(HelpersTestCase)


import numpy as np
import shapely.geometry
from shapely.affinity import translate

from gdshelpers.geometry.chip import Cell
from gdshelpers.helpers import StandardLayers
from gdshelpers.parts.text import Text
from gdshelpers.helpers.alignment import Alignment


class GridLayout(object):
    """
    A grid layout class.

    This class arranges cells in an ordered matter. It is useful if you want to place several independent devices
    next to each other.

    :param title: Title to be put on the top of the grid layout.
    :type title: str
    :param tight: Pack all devices as close as possible along the x-axis.
    :type tight: bool
    :param region_layer_type: One of ``None``, ``'layout'`` or ``cell``.
    :type region_layer_type: str
    :param region_layer_on_labels: Put a region layer around the labels as well when in ``cell`` mode.
    :type region_layer_on_labels: bool
    :param vertical_spacing: Minimum vertical spacing between devices. The true spacing will be bigger in most
                             cases, since it will align to the next write field.
    :type vertical_spacing: float
    :param vertical_alignment: Vertical alignment, typically the write field size.
    :type vertical_alignment: float
    :param horizontal_spacing: Minimum horizontal spacing between devices. The true spacing will be bigger in most
                               cases, since it will align to the next write field.
    :type horizontal_spacing: float
    :param horizontal_alignment: Horizontal alignment, typically the write field size.
    :type horizontal_alignment: float
    :param text_size: Size of the title text.
    :type text_size: float
    :param row_text_size: Size of text placed in the rows.
    :type row_text_size: float
    :param line_width: Width of the frame.
    :type line_width: float
    :param frame_layer: Layer of the frame. If set to zero, the frame will not be drawn.
    :type frame_layer: int
    :param text_layer: Layer of text
    :type text_layer: int
    :param region_layer: Layer of the region layer boxes.
    :type region_layer: int
    """

    def __init__(self, title=None, tight=False, region_layer_type='layout', region_layer_on_labels=False,
                 vertical_spacing=10, vertical_alignment=100, horizontal_spacing=10,
                 horizontal_alignment=100, text_size=40, row_text_size=30, line_width=1,
                 align_title_line=True,
                 frame_layer=StandardLayers.framelayer, text_layer=StandardLayers.patnamelayer,
                 region_layer=StandardLayers.regionlayer):
        self.title = title
        self.tight = tight
        self.vertical_spacing = vertical_spacing
        self.vertical_alignment = vertical_alignment
        self.horizontal_spacing = horizontal_spacing
        self.horizontal_alignment = horizontal_alignment
        self.text_size = text_size
        self.row_text_size = row_text_size
        self.line_width = line_width
        self.align_title_line = align_title_line
        self.frame_layer = frame_layer
        self.text_layer = text_layer
        self.region_layer = region_layer
        self.region_layer_type = region_layer_type
        self.region_layer_on_labels = region_layer_on_labels
        self._rows = list()
        self._current_row = None

    @property
    def region_layer_type(self):
        return self._region_layer_type

    # noinspection PyAttributeOutsideInit,PyShadowingBuiltins
    @region_layer_type.setter
    def region_layer_type(self, type):
        assert type in (None, 'layout', 'cell')
        self._region_layer_type = type

    def begin_new_row(self, row_label=None):
        """
        Begin a new row.

        :param row_label: Label of the row.
        :type row_label: str
        """
        self._finish_row()

        self._current_row = {'row_label': row_label,
                             'items': list()}

        self.add_label_to_row(row_label)

    def add_to_row(self, cell=None, bbox=None, alignment='left-bottom', realign=True, unique_id=None,
                   allow_region_layer=True):
        """
        Add a new cell to the row.

        The dimensions of the cell will either be determined directly or can be given via *bbox*. They determined
        bounding box always assumes a start at (0, 0) to preserve write field alignment.

        If *realign* is activated, the cell will be shifted, so that the it starts in the first write field quadrant.
        So if your cell contains structures as (-70, 10) and the write field size is 100, the cell will be shifted to
        (30, 10).

        The alignment of the cell can also be changed, but note that only 'left-bottom' guarantees write field
        alignment. When the region layers are used, these will force an alignment again, though.

        The position of each added cell inside the grid can be traced when passing a *unique_id* while adding the cell.

        :param cell: Cell to add.
        :type: gdsCAD.core.Cell
        :param bbox: Bounding box. If None, it will be determined based on *cell*.
        :param alignment: Alignment of the cell inside the grid.
        :param realign: Realign the cell, so that it starts in the first positive write field.
        :param unique_id: ID to trace where the cell was placed in the final layout.
        :param allow_region_layer: Allow a region layer around the cell.
        :type allow_region_layer: bool
        """
        assert self._current_row, 'Start a new row first'

        if cell is None:
            self._current_row['items'].append({'cell': None,
                                               'bbox': np.zeros([2, 2]),
                                               'offset': np.zeros(2),
                                               'id': unique_id,
                                               'alignment': Alignment(alignment),
                                               'allow_region_layer': False})
            return

        cell_bbox = np.reshape(cell.bounds, (2, 2)) if bbox is None else bbox

        if realign:
            offset = [self._remove_multiple_y_align(cell_bbox[0][i]) - cell_bbox[0][i] for i in (0, 1)]
            cell_bbox += offset
        else:
            offset = (0, 0)

        if bbox is None:
            cell_bbox[0, :] = 0

        bbox = cell_bbox

        self._current_row['items'].append({'cell': cell,
                                           'bbox': bbox,
                                           'offset': offset,
                                           'id': unique_id,
                                           'alignment': Alignment(alignment),
                                           'allow_region_layer': allow_region_layer})

    def add_label_to_row(self, text, size=None, origin=None, alignment='left-center'):
        """
        Add a label to the current row.

        :param text: Text of the label.
        :type text: str
        :param size: Size of the label. Defaults to *row_text_size*.
        :type size: float
        :param origin: Origin of the text.
        :param alignment: Alignment.
        """
        if text:
            size = size if size else self.row_text_size
            origin = origin if origin is not None else (0, 0)

            lab = Text(origin, size, text, true_bbox_alignment=True)
            elements = lab.get_shapely_object()
        else:
            elements = None
        self.add_to_row(elements, alignment=alignment, realign=False,
                        allow_region_layer=self.region_layer_on_labels)

    def add_column_label_row(self, labels, row_label=None, size=None, alignment='center-top'):
        """
        Start a new row, containing only labels.

        :param labels: List of labels for each column.
        :type labels: tuple, list
        :param row_label: Row label of the new row.
        :type: str, None
        :param size: Size of the text.
        :type size: float
        :param alignment: Alignment of the labels.
        """
        size = size if size else self.row_text_size

        self.begin_new_row(row_label=row_label)

        if isinstance(labels, str):
            labels = (labels,)

        for label in labels:
            self.add_label_to_row(label, size=size, alignment=alignment)

    def _finish_row(self):
        if self._current_row:
            self._rows.append(self._current_row)
            self._current_row = None

    def _remove_multiple_y_align(self, y):
        y %= self.vertical_alignment
        y = y + self.vertical_alignment if y < 0 else y
        return y

    def _remove_multiple_x_align(self, x):
        x %= self.horizontal_alignment
        x = x + self.horizontal_alignment if x < 0 else x
        return x

    def _next_y_align(self, y):
        if np.isclose(y % self.vertical_alignment, 0):
            return y

        if self.vertical_alignment:
            return (y // self.vertical_alignment + 1) * self.vertical_alignment
        else:
            return y

    def _next_x_align(self, x):
        if np.isclose(x % self.horizontal_alignment, 0):
            return x

        if self.horizontal_alignment:
            return (x // self.horizontal_alignment + 1) * self.horizontal_alignment
        else:
            return x

    def generate_layout(self):
        """
        Generate a layout cell.

        :return: Tuple of a cell, containing the layout and a dictionary mapping each unique id to the position inside
                 the cell.
        """
        self._finish_row()

        max_columns = 0
        column_properties = dict()
        row_properties = dict()
        # Find limits
        for row_id, row_dict in enumerate(self._rows):
            row_properties[row_id] = {'max_height': 0}

            for column_id, item in enumerate(row_dict['items']):
                max_columns = max(column_id, max_columns)

                if column_id not in column_properties:
                    column_properties[column_id] = {'max_width': 0}

                column_properties[column_id]['max_width'] = max(column_properties[column_id]['max_width'],
                                                                item['bbox'][1][0])
                row_properties[row_id]['max_height'] = max(row_properties[row_id]['max_height'],
                                                           item['bbox'][1][1])

        layout_cell = Cell('GRID_LAYOUT')
        pos = [0, self.vertical_spacing]
        limits = [0., 0.]
        mapping = dict()
        for row_id, row_dict in enumerate(self._rows):
            pos[0] = self.horizontal_spacing
            pos[1] = self._next_y_align(pos[1])

            max_height = row_properties[row_id]['max_height']
            for column_id, item in enumerate(row_dict['items']):
                max_width = column_properties[column_id]['max_width'] if not self.tight else item['bbox'][1][0]
                free_space_box = np.array(((0, 0), (max_width, max_height))) + pos

                offset = (item['alignment'].calculate_offset(item['bbox'])
                          - item['alignment'].calculate_offset(free_space_box))

                origin = offset + item['offset']

                item_unique_id = item['id']
                if item_unique_id is not None:
                    assert item_unique_id not in mapping, 'Recurring cell id, use unique values!'
                    mapping[item_unique_id] = origin

                if self.region_layer_type == 'cell' and item['allow_region_layer']:
                    # rl_box = shapely.geometry.box(pos[0], pos[1],
                    #                               self._next_x_align(pos[0] + max_width),
                    #                               self._next_y_align(pos[1] + max_height))
                    new_bbox = item['bbox'] + offset
                    delta = new_bbox[1, :] - new_bbox[0, :]
                    delta_x = self._next_x_align(delta[0])
                    delta_y = self._next_y_align(delta[1])

                    rl_box = shapely.geometry.box(new_bbox[0][0], new_bbox[0][1],
                                                  new_bbox[0][0] + delta_x, new_bbox[0][1] + delta_y)
                    layout_cell.add_to_layer(self.region_layer, rl_box)

                if item['cell']:
                    if isinstance(item['cell'], Cell):
                        layout_cell.add_cell(item['cell'], origin=origin)
                    else:
                        layout_cell.add_to_layer(self.text_layer, translate(item['cell'], *origin))

                next_x_pos = pos[0] + max_width + self.horizontal_spacing
                limits[0] = max(next_x_pos, limits[0])
                pos[0] = self._next_x_align(next_x_pos)

            next_y_pos = pos[1] + max_height + self.vertical_spacing
            limits[1] = max(next_y_pos, limits[1])
            pos[1] = next_y_pos

        if self.title:
            # If there is enough space for the title text until next alignment, use it
            tmp_text_obj = Text([0, 0], self.text_size, self.title).get_shapely_object()
            title_vertical_space = (tmp_text_obj.bounds[3] + 0.7 * self.text_size) + self.line_width
            title_horizontal_space = (tmp_text_obj.bounds[2] + self.text_size) + self.line_width
            limits[0] = max(limits[0], title_horizontal_space)

            if self.align_title_line:
                title_vertical_space = self._next_y_align(title_vertical_space) - self.line_width / 2.

            if (self._next_y_align(pos[1]) - limits[1]) >= title_vertical_space:
                limits[1] = pos[1] - title_vertical_space
            else:
                pos[1] = self._next_y_align(limits[1] + title_vertical_space)
                limits[1] = pos[1] - title_vertical_space

            # Paint vertical line
            if self.frame_layer:
                line = shapely.geometry.LineString([(self.line_width, limits[1]),
                                                    (self._next_x_align(limits[0]) - self.line_width, limits[1])])
                line = line.buffer(self.line_width)
                layout_cell.add_to_layer(self.frame_layer, line)

            title = Text([self.horizontal_spacing, (pos[1] + limits[1]) / 2.],
                         self.text_size, self.title, alignment='left-center')
            layout_cell.add_to_layer(self.text_layer, title)
        else:
            pos[1] = self._next_y_align(pos[1] + self.line_width)

        # Draw the frame
        if self.frame_layer:
            frame = shapely.geometry.box(0, 0, self._next_x_align(limits[0]), pos[1])
            frame = frame.difference(frame.buffer(-self.line_width))
            layout_cell.add_to_layer(self.frame_layer, frame)

        if self.region_layer_type == 'layout':
            frame = shapely.geometry.box(0, 0, self._next_x_align(limits[0]), pos[1])
            layout_cell.add_to_layer(self.region_layer, frame)

        return layout_cell, mapping


# noinspection PyPep8
def _example():
    layout = GridLayout('Test layout\nwith very cool features' + 'long! ' * 50, tight=False, region_layer_type=None)

    layout.add_column_label_row(['c0', 'c1'], 'col_labels')

    layout.begin_new_row('row1')

    test_cell_1 = Cell('TC1')
    test_cell_1.add_to_layer(1, shapely.geometry.box(0, 00, 335, 125))

    test_cell_2 = Cell('TC2')
    # test_cell_2.add_to_layer(1, shapely.geometry.box(101, 101, 335, 125))
    test_cell_2.add_to_layer(1, shapely.geometry.box(0, 0, 600, 400))

    layout.add_to_row(test_cell_1, realign=False)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_1, realign=True)  # , alignment='center-center')

    layout.begin_new_row('row2')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')

    layout.begin_new_row('row3')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')
    layout.add_to_row(test_cell_2, realign=True)  # , alignment='left-bottom')

    # layout.add_to_row(test_cell_1, realign=True)#, alignment='left-bottom')

    layout_cell, mapping = layout.generate_layout()

    layout_cell.show()
    layout_cell.save()


if __name__ == '__main__':
    _example()



import numpy as np
import shapely.geometry
import shapely.affinity
from shapely.geometry import Polygon, Point

from gdshelpers.geometry import geometric_union
from gdshelpers.parts import Port
from gdshelpers.parts.waveguide import Waveguide
from gdshelpers.geometry.chip import Cell

class SNSPD(object):
    def __init__(self, origin, angle, width, nw_width, nw_gap, nw_length, wing_span, wing_height, electrodes_pitch,
                 electrodes_gap, electrodes_height, waveguide_tapering, n_points=128):
        assert nw_gap > 0., 'Nanowire gap must be a positive value'
        assert nw_length > 0., 'Nanowire length must be a positive value'
        assert nw_width > 0., 'Nanowire width must be a positive value'

        # Let's directly create a port object. This simplifies later creation of the geometry
        # as well as checking the user parameters while creating the port.
        self._origin_port = Port(origin, angle, width)

        # Remember all the other stuff we got
        self.nw_gap = nw_gap
        self.nw_width = nw_width
        self.points = n_points
        self.nw_length = nw_length
        self.wing_span = wing_span
        self.wing_height = wing_height
        self.electrodes_pitch = electrodes_pitch
        self.electrodes_gap = electrodes_gap
        self.electrodes_height = electrodes_height
        self.waveguide_tapering = waveguide_tapering
        self.make_nanowire()
        self.make_waveguide()
        self.make_electrodes()

    @classmethod
    def make_at_port(cls, port, nw_width, nw_gap, nw_length, wing_span, wing_height, electrodes_pitch, electrodes_gap,
                     electrodes_height, waveguide_tapering, *kwargs):
        default_port_param = dict(port.get_parameters())
        default_port_param.update(kwargs)
        del default_port_param['origin']
        del default_port_param['angle']
        del default_port_param['width']
        return cls(port.origin, port.angle, port.width, nw_width, nw_gap, nw_length, wing_span, wing_height,
                   electrodes_pitch, electrodes_gap, electrodes_height, waveguide_tapering, **default_port_param)

    @staticmethod
    def _cub_bezier(origin, destination, aux_orig, aux_dest, n_points=200):
        x0, y0 = origin[0], origin[1]
        x1, y1 = aux_orig[0], aux_orig[1]
        x2, y2 = aux_dest[0], aux_dest[1]
        x3, y3 = destination[0], destination[1]

        a = x3 - 3 * x2 + 3 * x1 - x0
        b = 3 * x2 - 6 * x1 + 3 * x0
        c = 3 * x1 - 3 * x0
        d = x0

        e = y3 - 3 * y2 + 3 * y1 - y0
        f = 3 * y2 - 6 * y1 + 3 * y0
        g = 3 * y1 - 3 * y0
        h = y0

        points = []
        for t in range(n_points):
            t = t / float(n_points - 1)
            x = a * (t ** 3) + b * (t ** 2) + c * t + d
            y = e * (t ** 3) + f * (t ** 2) + g * t + h
            points.append([x, y])
        return points

    def get_shapely_object(self):
        return self.nw

    def make_nanowire(self):
        # U-shaped nanowire
        self.nanowire_port = Port(origin=(0, 0), angle=0, width=self.nw_width)
        nw = Waveguide.make_at_port(self.nanowire_port, width=self.nw_width)
        tip_radius = 0.5 * self.nw_gap + 0.5 * self.nw_width
        nw.add_bend(0.5 * np.pi, tip_radius)
        nw.add_straight_segment(self.nw_length - tip_radius)

        # Wing Square Pads which will overlap with the contact pads
        wing_tapering_origin = np.array(nw.current_port.origin)
        self.wing_pad_bottom_left = wing_tapering_origin + (
            (0.5 * self.wing_span - (self.nw_width + 0.5 * self.nw_gap)), 0.5 * self._origin_port.width)
        self.wing_pad_bottom_right = self.wing_pad_bottom_left + (0.5 * self.wing_span, 0)
        self.wing_pad_top_right = self.wing_pad_bottom_right + (0, self.wing_height)
        self.wing_pad_top_left = self.wing_pad_bottom_left + (0, self.wing_height)
        wing_pad = Polygon(
            [self.wing_pad_bottom_left, self.wing_pad_bottom_right, self.wing_pad_top_right, self.wing_pad_top_left])

        # Bezier tapering from nanowire to wing pads
        nw_left_side_coord = wing_tapering_origin - (0.5 * self.nw_width, 0)
        nw_right_side_coord = wing_tapering_origin + (0.5 * self.nw_width, 0)

        self.aux_origin_top_line = nw_left_side_coord + np.array(
            [0, 0.2 * (self.wing_pad_top_left[1] - nw_left_side_coord[1])])
        self.aux_dest_top_line = nw_left_side_coord + np.array(
            [0.8 * (self.wing_pad_top_left[0] - nw_left_side_coord[0]),
             0.4 * (self.wing_pad_top_left[1] - nw_left_side_coord[1])])

        wing_tapering_top_line = self._cub_bezier(nw_left_side_coord, self.wing_pad_top_left, self.aux_origin_top_line,
                                                  self.aux_dest_top_line)

        self.aux_origin_bottom_line = nw_right_side_coord + np.array(
            [0.1 * (self.wing_pad_bottom_left[0] - nw_right_side_coord[0]),
             self.wing_pad_bottom_left[1] - nw_right_side_coord[1]])
        self.aux_dest_bottom_line = nw_right_side_coord + np.array(
            [0, 0.5 * (self.wing_pad_bottom_left[1] - nw_right_side_coord[1])])
        wing_tapering_bottom_line = self._cub_bezier(self.wing_pad_bottom_left, nw_right_side_coord,
                                                     self.aux_origin_bottom_line, self.aux_dest_bottom_line)

        wing_tapering = Polygon(wing_tapering_top_line + wing_tapering_bottom_line)

        wing = geometric_union([wing_pad, wing_tapering])

        nw = geometric_union([nw, wing])
        nw_l = shapely.affinity.scale(nw, xfact=-1.0, yfact=1.0, zfact=1.0,
                                      origin=[self.nanowire_port.origin[0], self.nanowire_port.origin[1], 0])
        nw = geometric_union([nw, nw_l])
        nw = shapely.affinity.translate(nw, yoff=0.5 * self.nw_width)
        nw = shapely.affinity.rotate(nw, self._origin_port.angle - 0.5 * np.pi, origin=[0, 0], use_radians=True)
        self.nw = shapely.affinity.translate(nw, xoff=self._origin_port.origin[0], yoff=self._origin_port.origin[1])

        re_origin = self._origin_port.origin
        re_angle = self._origin_port.angle - 0.5 * np.pi
        rep = Port(origin=re_origin, angle=re_angle, width=self.wing_height)
        self.rep = rep.longitudinal_offset(self.wing_pad_bottom_left[0] + 0.25 * self.wing_span).parallel_offset(
            self.wing_pad_bottom_left[1] + 0.5 * self.wing_height)
        self.lep = self.rep.longitudinal_offset(-1.5 * self.wing_span).rotated(np.pi)

    def make_waveguide(self):
        wg = Waveguide.make_at_port(self._origin_port)
        wg.add_straight_segment(self.nw_length + 0.5 * self.nw_width + self.wing_height)
        if self.waveguide_tapering:
            wg.add_straight_segment(2. * self._origin_port.width, final_width=0.01)
        self.waveguide_port = wg.current_port
        wg = geometric_union([wg])
        nw = self.nw.difference(wg)
        buffer = nw.buffer(.2)
        self.wg = geometric_union([wg, buffer])

    def make_electrodes(self):
        bottom_left = self.nanowire_port.origin + np.array([0.5 * self.electrodes_gap, self.nw_length + 25.])
        top_left = bottom_left + (0, self.electrodes_height)
        electrode_width = self.electrodes_pitch - self.electrodes_gap
        top_right = top_left + (electrode_width, 0)
        bottom_right = (top_right[0], bottom_left[1])
        bottom_middle = np.array((bottom_right[0] - 0.5 * electrode_width, self.wing_pad_bottom_left[1]))
        pad = Polygon(
            [self.wing_pad_bottom_left, self.wing_pad_top_left, self.wing_pad_top_right, bottom_left, top_left,
             top_right, bottom_right, bottom_middle])
        pad = geometric_union([pad])
        pad = shapely.affinity.translate(pad, yoff=0.5 * self.nw_width)
        pad_l = shapely.affinity.scale(pad, xfact=-1.0, yfact=1.0, zfact=1.0, origin=[0, 0, 0])
        pad = geometric_union([pad, pad_l])

        pad = shapely.affinity.rotate(pad, self._origin_port.angle - 0.5 * np.pi,
                                      origin=[0, 0],
                                      use_radians=True)
        self.pad = shapely.affinity.translate(pad, xoff=self._origin_port.origin[0], yoff=self._origin_port.origin[1])

    def get_waveguide(self):
        return self.wg

    def get_electrodes(self):
        return self.pad

    def get_passivation_layer(self, passivation_buffer=0.1):
        buffer = self.nw.buffer(passivation_buffer)
        passivation_layer = geometric_union([buffer])
        return passivation_layer

    def get_aux_top(self):
        return self.aux_origin_top_line, self.aux_dest_top_line

    def get_aux_bottom(self):
        return self.aux_origin_bottom_line, self.aux_dest_bottom_line

    @property
    def right_electrode_port(self):
        return self.rep

    @property
    def left_electrode_port(self):
        return self.lep

    @property
    def current_port(self):
        return self.waveguide_port


snspd_parameters = {
    'nw_width': 0.1,
    'nw_gap': 0.1,
    'nw_length': 70.,
    'wing_span': 10.,
    'wing_height': 5.,
    'electrodes_pitch': 125.,
    'electrodes_gap': 45.,
    'electrodes_height': 600.,
    'waveguide_tapering': False,
}


def example():
    start_port = Port((0, 0), 0.5 * np.pi, 1.)
    wg1 = Waveguide.make_at_port(start_port)
    wg1.add_straight_segment(100.)
    # wg1.add_bend(0, 60.)
    detector = SNSPD.make_at_port(wg1.current_port, **snspd_parameters)
    wg = detector.get_waveguide()
    electrodes = detector.get_electrodes()
    pl = detector.get_passivation_layer(passivation_buffer=0.2)

    wg2 = Waveguide.make_at_port(detector.current_port)
    wg2.add_straight_segment(50.)
    # cell = gdsCAD.core.Cell('test')
    cell = Cell('test')
    cell.add_to_layer(3, Point(detector.get_aux_top()[0]).buffer(0.05))
    cell.add_to_layer(4, Point(detector.get_aux_top()[1]).buffer(0.05))
    cell.add_to_layer(3, Point(detector.get_aux_bottom()[0]).buffer(0.05))
    cell.add_to_layer(4, Point(detector.get_aux_bottom()[1]).buffer(0.05))
    cell.add_to_layer(3, wg1)

    cell.add_to_layer(1, detector)
    cell.add_to_layer(2, detector.get_waveguide())
    cell.add_to_layer(6, detector.get_electrodes())
    cell.add_to_layer(5, pl)

    # cell.add_to_layer(6, detector.right_electrode_port.debug_shape)
    # cell.add_to_layer(6, detector.left_electrode_port.debug_shape)
    cell.add_to_layer(3, wg2)
    cell.save('SNSPD_test.gds')
    cell.show()


if __name__ == '__main__':
    example()



from __future__ import print_function, division

import collections

import shapely.geometry
import shapely.affinity
import shapely.ops
import shapely.validation

from gdshelpers.parts import Port
from gdshelpers.helpers import find_line_intersection, normalize_phase
from gdshelpers.helpers.bezier import CubicBezierCurve
from gdshelpers.helpers.positive_resist import convert_to_positive_resist

import numpy as np
import numpy.linalg as linalg
import scipy.interpolate


class Waveguide(object):
    def __init__(self, origin, angle, width):
        self._current_port = Port(origin, angle, width)
        self._in_port = self._current_port.inverted_direction.copy()
        self._segments = list()

    @classmethod
    def make_at_port(cls, port, **kargs):
        port_param = port.copy()
        port_param.set_port_properties(**kargs)
        return cls(**port_param.get_parameters())

    @property
    def x(self):
        return self._current_port.origin[0]

    @property
    def y(self):
        return self._current_port.origin[1]

    @property
    def origin(self):
        return self._current_port.origin

    @property
    def angle(self):
        return self._current_port.angle

    @property
    def width(self):
        return self._current_port.width

    @width.setter
    def width(self, width):
        self._current_port.width = width

    @property
    def current_port(self):
        return self._current_port.copy()

    # Add alias for current port
    port = current_port

    @property
    def in_port(self):
        return self._in_port.copy()

    @property
    def length(self):
        return sum((length for port, obj, length in self._segments))

    @property
    def length_last_segment(self):
        if not len(self._segments):
            return 0
        return self._segments[-1][2]

    def get_shapely_object(self):
        """
        Get a shapely object which forms this path.
        """
        return shapely.ops.cascaded_union([obj for port, obj, length in self._segments])

    def get_segments(self):
        """
        Returns the list of tuples, containing their ports and shapely objects.
        """
        return [(port.copy(), obj, length) for port, obj, length in self._segments]

    def add_straight_segment(self, length, final_width=None, **kargs):
        self._current_port.set_port_properties(**kargs)
        final_width = final_width or self.width

        endpoint = shapely.geometry.Point(length, 0)

        if not np.isclose(length, 0):
            assert length >= 0, 'Length of straight segment must not be negative'
            polygon = shapely.geometry.Polygon(([0, self.width/2], [length, final_width/2],
                                                [length, -final_width/2], [0, -self.width/2]))
            polygon = shapely.affinity.rotate(polygon, self.angle, origin=[0, 0], use_radians=True)
            polygon = shapely.affinity.translate(polygon, self.x, self.y)
            self._segments.append((self._current_port.copy(), polygon, length))

        endpoint = shapely.affinity.rotate(endpoint, self.angle, origin=[0, 0], use_radians=True)
        endpoint = shapely.affinity.translate(endpoint, self.x, self.y)

        self._current_port.origin = endpoint.coords[0]
        self._current_port.width = final_width
        return self

    def add_arc(self, final_angle, radius, final_width=None, n_points=128, shortest=True, **kwargs):
        delta = final_angle-self.angle
        if not np.isclose(normalize_phase(delta), 0):
            if shortest:
                delta = normalize_phase(delta)
            self.add_bend(delta, radius, final_width, n_points, **kwargs)
        return self

    def add_bend(self, angle, radius, final_width=None, n_points=128, **kargs):
        # If number of points is None, default to 128
        n_points = n_points if n_points else 128

        self._current_port.set_port_properties(**kargs)
        final_width = final_width or self.width

        angle = normalize_phase(angle, zero_to_two_pi=True) - (0 if angle > 0 else 2*np.pi)

        if not np.isclose(radius, 0) and not np.isclose(angle, 0) and radius > 0:
            if angle > 0:
                circle_center = (-np.sin(self.angle)*radius, np.cos(self.angle)*radius) + self.current_port.origin
                start_angle = -np.pi/2 + self.angle
            else:
                circle_center = (np.sin(self.angle)*radius, -np.cos(self.angle)*radius) + self.current_port.origin
                start_angle = np.pi/2 + self.angle

            end_angle = start_angle + angle

            # Calculate the points needed for this angle
            points = max(abs(end_angle - start_angle)/(np.pi/2) * n_points, 2)

            phi = np.linspace(start_angle, end_angle, points)
            upper_radius_points = np.linspace(radius - self.width/2, radius-final_width/2, points)
            upper_line_points = np.array([upper_radius_points*np.cos(phi),
                                          upper_radius_points*np.sin(phi)]).T + circle_center

            lower_radius_points = np.linspace(radius + self.width/2, radius+final_width/2, points)
            lower_line_points = np.array([lower_radius_points*np.cos(phi),
                                          lower_radius_points*np.sin(phi)]).T + circle_center

            polygon = shapely.geometry.Polygon(np.concatenate([upper_line_points, lower_line_points[::-1, :]]))
            self._segments.append((self._current_port.copy(), polygon, abs(angle)*radius))

            endpoint = shapely.geometry.Point(radius * np.cos(end_angle) + circle_center[0],
                                              radius * np.sin(end_angle) + circle_center[1])
            self._current_port.origin = endpoint.coords[0]

        self._current_port.width = final_width
        self._current_port.angle += angle

        # assert self._segments[-1][1].is_valid, \
        #     'Invalid polygon generated: %s' % shapely.validation.explain_validity(self._segments[-1][1])

        return self

    def add_parameterized_path(self, path, width=None, sample_distance=0.50, sample_points=100, path_derivative=None,
                               path_function_supports_numpy=False, width_function_supports_numpy=False):
        """
        Generate a parameterized path.

        The path coordinate system is the origin and rotation of the current path. So if you want to continue your path
        start at (0, 0) in y-direction.

        Note, that path is either a list of (x,y) coordinate tuples or a callable function which takes one float
        parameter between 0 and 1. If you use a parameterized function, its first derivative must be continuous.
        When using a list of coordinates, these points will be connected by straight lines. They must be sufficiently
        close together to simulate a first derivative continuous path.

        This function will try to space the final points of the curve equidistantly. To achieve this, it will first
        sample the function and find its first derivative. Afterwards it can calculate the cumulative sum of the length
        of the first derivative. This allows to sample the function nearly equidistantly in a second step. This
        approach might be wasteful for paths like (x**2, y). You can suppress resampling for length by passing zero or
        none as sample_distance parameter.

        The width of the generated waveguide may either be constant when passing a number or also be a callable
        function, using the same parameter as the path.

        Note, that your final direction of the path might not be what you expected. This is caused by the numerical
        procedure which generates numerical errors when calculating the first derivative. You can either append another
        arc to the waveguide to get to you a correct angle or you can also supply a function which is the algebraic
        first derivative. The returned vector is not required to be normed.

        By default, for each parameter point t, the parameterized functions are call. You will notice that this is
        rather slow. To achieve the best performance, write your functions in such a way, that they can handle a
        numpy array as parameter *t*. Once the *path_function_supports_numpy* option is set to True, the function will
        be called only once, speeding up the calculation considerable.

        :param path:
        :param width:
        :param sample_distance:
        :param sample_points:
        :param path_derivative:
        :param path_function_supports_numpy:
        :param width_function_supports_numpy:
        """

        path_callable = callable(path)
        if path_callable:
            presample_t = np.linspace(0, 1, sample_points)

            if path_function_supports_numpy:
                presample_coordinates = np.array(path(presample_t)).T
            else:
                presample_coordinates = np.array([path(x) for x in presample_t])

            if sample_distance:
                # # Calculate the derivative
                # if path_derivative:
                #     assert callable(path_derivative), 'The derivative of the path function must be callable'
                #     presample_coordinates_d1 = np.array([path_derivative(x) for x in presample_t[:-1]])
                # else:
                #     presample_coordinates_d1 = np.diff(presample_coordinates, axis=0)
                presample_coordinates_d1 = np.diff(presample_coordinates, axis=0)
                presample_coordinates_d1_norm = np.apply_along_axis(linalg.norm, 1, presample_coordinates_d1)
                presample_coordinates_d1__cum_norm = np.insert(np.cumsum(presample_coordinates_d1_norm), 0, 0)

                lengths = np.linspace(presample_coordinates_d1__cum_norm[0],
                                      presample_coordinates_d1__cum_norm[-1],
                                      presample_coordinates_d1__cum_norm[-1]//sample_distance)

                # First get the spline representation. This is needed since we manipulate these directly for roots
                # finding.
                spline_rep = scipy.interpolate.splrep(presample_t, presample_coordinates_d1__cum_norm, s=0)

                def find_y(y):
                    interp_result = scipy.interpolate.sproot((spline_rep[0], spline_rep[1] - y, spline_rep[2]), mest=1)
                    return interp_result[0] if len(interp_result) else None

                # We need a small hack here and exclude lengths[0]==0 since it finds no root there
                sample_t = np.array([0, ] + [find_y(length) for length in lengths[1:-1]] + [1, ])

                if path_function_supports_numpy:
                    sample_coordinates = np.array(path(sample_t)).T
                else:
                    sample_coordinates = np.array([path(x) for x in sample_t])
            else:
                sample_coordinates = presample_coordinates
                sample_t = presample_t
        else:
            # If we do not have a sample function, we need to "invent a sampling parameter"
            sample_coordinates = np.array(path)
            sample_t = np.linspace(0, 1, sample_coordinates.shape[0])

        # Calculate the derivative
        if path_derivative:
            assert callable(path_derivative), 'The derivative of the path function must be callable'
            if path_function_supports_numpy:
                sample_coordinates_d1 = np.array(path_derivative(sample_t)).T
            else:
                sample_coordinates_d1 = np.array([path_derivative(x) for x in sample_t])
        else:
            sample_coordinates_d1 = np.vstack(([1, 0], np.diff(sample_coordinates, axis=0)))

        sample_coordinates_d1_norm = np.apply_along_axis(linalg.norm, 1, sample_coordinates_d1)
        sample_coordinates_d1_normed = sample_coordinates_d1 / sample_coordinates_d1_norm[:, None]

        # Find the orthogonal vectors to the derivative
        sample_coordinates_d1_normed_ortho = np.vstack((sample_coordinates_d1_normed[:, 1],
                                                        -sample_coordinates_d1_normed[:, 0])).T

        # Calculate the width of the waveguide at the given positions
        if callable(width):
            if width_function_supports_numpy:
                sample_width = width(sample_t)
            else:
                sample_width = np.array([width(x) for x in sample_t])
        else:
            sample_width = np.array([(width if width else self.current_port.width), ])

        # Now we have everything to calculate the polygon
        poly_path_1 = sample_coordinates + sample_width[:, None]/2 * sample_coordinates_d1_normed_ortho
        poly_path_2 = sample_coordinates - sample_width[:, None]/2 * sample_coordinates_d1_normed_ortho

        assert shapely.geometry.LineString(np.concatenate([poly_path_1, poly_path_2[::-1, :]])).is_simple, \
            'Outer lines of parameterized wg intersect. Try using lower bend radii or smaller a smaller wg'

        # Now add the shapely objects and do book keeping
        polygon = shapely.geometry.Polygon(np.concatenate([poly_path_1, poly_path_2[::-1, :]]))
        assert polygon.is_valid, 'Generated polygon path is not valid: %s' % \
                                 shapely.validation.explain_validity(polygon)

        polygon = shapely.affinity.rotate(polygon, self.angle, origin=[0, 0], use_radians=True)
        polygon = shapely.affinity.translate(polygon, self.x, self.y)

        length = np.sum(np.apply_along_axis(linalg.norm, 1, np.diff(sample_coordinates, axis=0)))
        self._segments.append((self._current_port.copy(), polygon, length))

        endpoint = shapely.geometry.Point(sample_coordinates[-1][0], sample_coordinates[-1][1])
        endpoint = shapely.affinity.rotate(endpoint, self.angle, origin=[0, 0], use_radians=True)
        endpoint = shapely.affinity.translate(endpoint, self.x, self.y)
        self._current_port.origin = endpoint.coords[0]

        self._current_port.width = sample_width[-1] if callable(width) else sample_width
        self._current_port.angle += np.arctan2(sample_coordinates_d1[-1][1], sample_coordinates_d1[-1][0])
        return self

    def add_cubic_bezier_path(self, p0, p1, p2, p3, width=None, **kwargs):
        """
        Add a cubic bezier path to the waveguide.

        Coordinates are in the "waveguide tip coordinate system", so the first point will probably be p0 == (0, 0).
        Note that your bezier curve undergoes the same restrictions as a parameterized path. Don't self-intersect it and
        don't use small bend radii.

        :param p0: 2 element tuple like coordinates
        :param p1: 2 element tuple like coordinates
        :param p2: 2 element tuple like coordinates
        :param p3: 2 element tuple like coordinates
        :param width: Width of the waveguide, as passed to :func:add_parameterized_path
        :param kwargs: Optional keyword arguments, passed to :func:add_parameterized_path
        :return: Changed waveguide
        :rtype: Waveguide
        """

        bezier_curve = CubicBezierCurve(p0, p1, p2, p3)

        self.add_parameterized_path(path=bezier_curve.evaluate, path_derivative=bezier_curve.evaluate_d1, width=width,
                                    path_function_supports_numpy=True,  **kwargs)
        return self

    def add_bezier_to(self, final_coordinates, final_angle, bend_strength, width=None, **kwargs):

        try:
            bs1, bs2 = float(bend_strength[0]), float(bend_strength[1])
        except (KeyError, TypeError):
            bs1 = bs2 = float(bend_strength)

        final_port = Port(final_coordinates, final_angle, self.width)
        p0 = (0, 0)
        p1 = self.current_port.longitudinal_offset(bs1).origin - self.current_port.origin
        p2 = final_port.longitudinal_offset(-bs2).origin - self.current_port.origin
        p3 = final_coordinates - self.current_port.origin

        tmp_wg = Waveguide.make_at_port(self.current_port.copy().set_port_properties(angle=0))
        tmp_wg.add_cubic_bezier_path(p0, p1, p2, p3, width=width, **kwargs)

        self._segments.append((self._current_port.copy(), tmp_wg.get_shapely_object(), tmp_wg.length))
        self._current_port = tmp_wg.current_port

        return self

    def add_bezier_to_port(self, port, bend_strength, width=None, **kwargs):
        if not width and not np.isclose(self.width, port.width):
            width = lambda t: t*(port.width-self.width) + self.width
            supports_numpy = True
        else:
            supports_numpy = False

        kwargs['width_function_supports_numpy'] = kwargs.get('width_function_supports_numpy', supports_numpy)

        self.add_bezier_to(port.origin, port.inverted_direction.angle, bend_strength, width, **kwargs)
        return self

    def add_route_single_circle_to(self, final_coordinates, final_angle, max_bend_strength=None,
                                   on_line_only=False):

        """
        Connect two points by straight lines and one circle.

        Works for geometries like round edges and others. The final straight line can also be omitted so that
        the waveguide only end on the line described by the support vector and angle.

        By default, this method tries to route to the target with the greatest possible circle. But the valid
        bending range may be limited via the `max_bend_strength` parameter.

        This method does not work for geometries which cannot be connected only by straight lines and one
        circle, such as parallel lines etc.

        Still, this method can prove extremely useful for routing to i.e. grating couplers etc.

        :param final_coordinates: Final destination point.
        :param final_angle: Final angle of the waveguide.
        :param max_bend_strength: The maximum allowed bending radius.
        :param on_line_only: Omit the last straight line and only route to described line.
        """
        # We are given an out angle, but the internal math is for inward pointing lines
        final_angle = normalize_phase(final_angle) + np.pi

        # We need to to some linear algebra. We first find the intersection of the two waveguides
        r1 = self.current_port.origin
        r2 = np.array(final_coordinates)
        intersection_point, distance = find_line_intersection(r1, self.angle, r2, final_angle)

        assert distance[0] >= 0, 'Origin waveguide is too close to target. No curve possible'
        assert on_line_only or distance[1] >= 0, 'Target position is too close to target. No curve possible'

        # Calculate the angle bisector
        u1 = np.array([np.cos(self.angle), np.sin(self.angle)])
        u2 = np.array([np.cos(final_angle), np.sin(final_angle)])
        u_half = u1+u2
        half_angle = np.arctan2(u_half[1], u_half[0])
        diff_angle = normalize_phase(self.angle-final_angle - np.pi)

        _, r1 = find_line_intersection(r1, self.angle + np.pi/2, intersection_point, half_angle)

        if not on_line_only:
            max_poss_radius = min(np.abs([r1[0], distance[0]/np.tan(diff_angle/2), distance[1]/np.tan(diff_angle/2)]))
        else:
            max_poss_radius = min(np.abs([r1[0], distance[0]/np.tan(diff_angle/2)]))

        radius = min([max_bend_strength, max_poss_radius]) if max_bend_strength is not None else max_poss_radius
        d = abs(radius*np.tan(diff_angle/2))

        tmp_wg = Waveguide.make_at_port(self.current_port)
        tmp_wg.add_straight_segment(distance[0] - d)
        tmp_wg.add_bend(-diff_angle, radius)

        self._segments.append((self._current_port.copy(), tmp_wg.get_shapely_object(), tmp_wg.length))
        self._current_port = tmp_wg.current_port

        if not on_line_only:
            self.add_straight_segment(distance[1] - d)

        return self

    def add_route_single_circle_to_port(self, port, max_bend_strength=None, on_line_only=False):
        """
        Connect to port by straight lines and one circle.

        Helper function to conveniently call add_route_single_circle_to.
        :param port: Target port.
        :param max_bend_strength: The maximum allowed bending radius.
        :param on_line_only: Omit the last straight line and only route to line described by port.
        """
        self.add_route_single_circle_to(port.origin, port.inverted_direction.angle, max_bend_strength, on_line_only)
        return self

    def add_straight_segment_to_intersection(self, line_origin, line_angle, **line_kw):
        """
        Add a straight line until it intersects with an other line.

        The other line is described by the support vector and the line angle.

        :param line_origin: Intersection line support vector.
        :param line_angle: Intersection line angle.
        :param line_kw: Parameters passed on to add_straight_segment.
        :raise ArithmeticError: When there is no intersection due to being parallel or if
                                the intersection is behind the waveguide.
        """
        r1 = self.current_port.origin
        r2 = np.array(line_origin)

        try:
            intersection_point, distance = find_line_intersection(r1, self.angle, r2, line_angle)
            if np.isclose(distance[0], 0):
                return self
            elif distance[0] < 0:
                raise ArithmeticError('No forward intersection of waveguide with given line')
            self.add_straight_segment(distance[0], **line_kw)
        except linalg.LinAlgError:
            # We do not find an intersection if lines are parallel
            raise ArithmeticError('There is no intersection with the given line')
        return self

    def add_straight_segment_until_x(self, x, **line_kw):
        """
        Add straight segment until the given x value is reached.

        :param x: value
        :param line_kw: Parameters passed on to add_straight_segment.
        """
        self.add_straight_segment_to_intersection([x, 0], np.pi/2, **line_kw)
        return self

    def add_straight_segment_until_y(self, y, **line_kw):
        """
        Add straight segment until the given y value is reached.

        :param y: value
        :param line_kw: Parameters passed on to add_straight_segment.
        """
        self.add_straight_segment_to_intersection([0, y], 0, **line_kw)
        return self

    def add_straight_segment_until_level_of_port(self, port, **line_kw):
        """
        Add a straight line until it is on the same level as the given port. If several ports are given in a list,
        the most distant port is chosen.

        In this context "on the same level" means the intersection of the waveguide with the line
        orthogonal to the given port.

        :param port: The port or a list of ports.
        :param line_kw:
        """

        if isinstance(port, collections.Iterable):
            direction = (np.cos(self.angle), np.sin(self.angle))
            distances = np.array([np.dot(p.origin, direction) for p in port])
            furthest_port_idx = distances.argmax()
            port = port[furthest_port_idx]

        self.add_straight_segment_to_intersection(port.origin, port.angle - np.pi/2, **line_kw)
        return self


def _example():
    import gdsCAD
    import gdsCAD.templates

    from gdshelpers.geometry import convert_to_gdscad, geometric_union
    from gdshelpers.parts.splitter import Splitter
    from gdshelpers.parts.coupler import GratingCoupler

    port = Port([0, 0], 0, 1)
    path = Waveguide.make_at_port(port)

    path.add_straight_segment(10)
    path.add_bend(np.pi/2, 10, final_width=0.5)
    path.add_bend(-np.pi/2, 10, final_width=1)
    path.add_arc(np.pi*3/4, 10,)

    splitter = Splitter.make_at_root_port(path.current_port, 30, 10)
    path2 = Waveguide.make_at_port(splitter.right_branch_port)
    path2.add_bend(-np.pi/4, 10)

    n = 10
    path2.add_parameterized_path(lambda t: (n*10*t, np.cos(n*2*np.pi*t)-1),
                                 path_derivative=lambda t: (n*10, -n*2*np.pi*np.sin(n*2*np.pi*t)),
                                 width=lambda t: np.cos(n*2*np.pi*t)*0.2+np.exp(-t)*0.3 + 0.5,
                                 width_function_supports_numpy=True)
    path2.add_straight_segment(10)
    print(path2.length)
    print(path2.length_last_segment)

    path2.add_cubic_bezier_path((0, 0), (5, 0), (10, 10), (5, 10))

    coupler1 = GratingCoupler([100, 50], 0, 1, np.deg2rad(30), [10, 0.1, 2, 0.1, 2], start_radius_absolute=True)

    path3 = Waveguide((0, -50), np.deg2rad(0), 1)
    path3.add_bezier_to(coupler1.port.origin, coupler1.port.inverted_direction.angle, bend_strength=50)

    splitter2 = Splitter.make_at_left_branch_port(splitter.left_branch_port, 30, 10, wavelength_root=2)

    path4 = Waveguide.make_at_port(splitter2.root_port)
    path4.add_straight_segment(20)
    path4.width = 1
    path4.add_straight_segment(20)

    empty_path = Waveguide.make_at_port(path4.current_port)

    whole_layout = (path, splitter, path2, splitter2, coupler1, path3, path4, empty_path)



    layout = gdsCAD.core.Layout('LIBRARY')
    cell = gdsCAD.core.Cell('TOP')
    cell.add(convert_to_gdscad(whole_layout))
    cell.add(convert_to_gdscad(empty_path, layer=2))
    cell.add(convert_to_gdscad(splitter.root_port.debug_shape, layer=4))

    layout.add(cell)


    cell_df = gdsCAD.core.Cell('TOP_DF')
    cell_df.add(convert_to_gdscad(convert_to_positive_resist(whole_layout, buffer_radius=1.5)))
    layout.add(cell_df)

    layout.save('output.gds')
    cell.show()

if __name__ == '__main__':
    _example()



from gdshelpers.geometry import geometric_union

import numpy as np


def convert_to_positive_resist(parts, buffer_radius, outer_resolution=None):
    """
    Convert a list of parts and shapely objects to a positive resist design by
    adding a buffer around the actual design.

    :param parts: List of parts and shapely objects.
    :param buffer_radius: Buffer radius
    :param outer_resolution: Outer buffer circumference resolution. Defaults to one 20th of the buffer radius.
    :return: Converted Shapely geometry.
    :rtype: shapely.base.BaseGeometry
    """
    outer_resolution = buffer_radius / 20. if outer_resolution is None else outer_resolution

    assert buffer_radius > 0, 'The buffer radius must be positive.'
    assert outer_resolution >= 0, 'Resolution must be positive or zero.'

    parts = (parts, ) if not isinstance(parts, (tuple, list)) else parts

    # First merge all parts into one big shapely object
    union = geometric_union(parts)

    # Sometimes those polygons do not touch correctly and have micro gaps due to
    # floating point precision. We work around this by inflating the object a tiny bit.
    fixed_union = union.buffer(np.finfo(np.float32).eps, resolution=0)

    # Generate the outer polygon and simplify if required
    outer_poly = union.buffer(buffer_radius)

    if outer_resolution:
        outer_poly = outer_poly.simplify(outer_resolution)

    # Substract the original parts from outer poly
    inverted = outer_poly.difference(fixed_union)

    return inverted


from .small import int_to_alphabet, id_to_alphanumeric, normalize_phase, find_line_intersection, \
    raith_eline_dosefactor_to_datatype
# noinspection PyPep8Naming
from . import layers as StandardLayers

__all__ = ['int_to_alphabet', 'id_to_alphanumeric', 'normalize_phase', 'find_line_intersection',
           'raith_eline_dosefactor_to_datatype', 'StandardLayers']



# -*- coding: utf-8 -*-
from setuptools import setup, find_packages
from os import path

import gdshelpers

with open(path.join(path.abspath(path.dirname(__file__)), 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='gdsHelpers',
    version=gdshelpers.__version__,
    author='Helge Gehring, Matthias Blaicher, Wladick Hartmann, Wolfram Pernice',
    author_email='helge.gehring@uni-muenster.de',
    url='https://github.com/HelgeGehring/gdshelpers',
    packages=find_packages(),
    platforms='All',
    python_requires='>=3',
    license='LGPLv3',
    description='A simple Python package for creating or reading GDSII/OASIS layout files.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    requires=['matplotlib', 'numpy', 'shapely', 'scipy'],
    extras_require={
        'gdspy_export': ['gdspy(>=1.3.1)'],
        'gdscad_export': ['gdscad'],
        'oasis_export': ['fatamorgana'],
        'image_export': ['descartes'],
        'mesh_export': ['trimesh']
    },
    test_suite='gdshelpers.tests.test_suite',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Electronic Design Automation (EDA)',
        'Topic :: Scientific/Engineering :: Physics',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)'
    ]
)
